"""Tests related to uaclient.entitlement.base module."""

import copy
import logging
from functools import partial

import mock
import pytest

from uaclient import exceptions, messages, system, util
from uaclient.entitlements import base
from uaclient.entitlements.entitlement_status import (
    ApplicabilityStatus,
    ApplicationStatus,
    CanDisableFailure,
    CanDisableFailureReason,
    CanEnableFailure,
    CanEnableFailureReason,
    UserFacingStatus,
)
from uaclient.status import ContractStatus


class ConcreteTestEntitlement(base.UAEntitlement):
    name = "testconcreteentitlement"
    title = "Test Concrete Entitlement"
    description = "Entitlement for testing"

    def __init__(
        self,
        cfg=None,
        disable=None,
        enable=None,
        applicability_status=None,
        application_status=(ApplicationStatus.DISABLED, ""),
        supports_access_only=False,
        access_only=False,
        supports_purge=False,
        purge=False,
        dependent_services=None,
        required_services=None,
        incompatible_services=None,
        blocking_incompatible_services=None,
        variant_name="",
        **kwargs
    ):
        super().__init__(
            cfg,
            access_only=access_only,
            purge=purge,
        )
        self.supports_access_only = supports_access_only
        self.supports_purge = supports_purge
        self._disable = disable
        self._enable = enable
        self._applicability_status = applicability_status
        self._application_status = application_status
        self._dependent_services = dependent_services
        self._required_services = required_services
        self._incompatible_services = incompatible_services
        self._blocking_incompatible_services = blocking_incompatible_services
        self._variant_name = variant_name

    def _perform_disable(self, *args, **kwargs):
        self._application_status = (
            ApplicationStatus.DISABLED,
            "disable() called",
        )
        return self._disable

    def _perform_enable(self, *args, **kwargs):
        return self._enable

    def enable_steps(self):
        return 1

    def disable_steps(self):
        return 1

    def applicability_status(self):
        if self._applicability_status is not None:
            return self._applicability_status
        else:
            return super().applicability_status()

    def application_status(self):
        return self._application_status

    def blocking_incompatible_services(self):
        if self._blocking_incompatible_services is not None:
            return self._blocking_incompatible_services
        else:
            return super().blocking_incompatible_services()

    @property
    def variant_name(self):
        return self._variant_name

    @property
    def is_variant(self):
        return False if not self._variant_name else True


@pytest.fixture
def base_entitlement_factory(entitlement_factory):
    return partial(entitlement_factory, ConcreteTestEntitlement)


class TestEntitlement:
    def test_entitlement_abstract_class(self):
        """UAEntitlement is abstract requiring concrete methods."""
        with pytest.raises(TypeError):
            base.UAEntitlement()

    def test_init_default_sets_up_uaconfig(self):
        """UAEntitlement sets up a uaconfig instance upon init."""
        entitlement = ConcreteTestEntitlement()
        assert "/var/lib/ubuntu-advantage" == entitlement.cfg.data_dir

    def test_init_accepts_a_uaconfig(self, FakeConfig):
        """An instance of UAConfig can be passed to UAEntitlement."""
        cfg = FakeConfig(cfg_overrides={"data_dir": "/some/path"})
        entitlement = ConcreteTestEntitlement(cfg)
        assert "/some/path" == entitlement.cfg.data_dir


class TestEntitlementNames:
    @pytest.mark.parametrize(
        "p_name,expected",
        (
            ("pretty_name", ["testconcreteentitlement", "pretty_name"]),
            ("testconcreteentitlement", ["testconcreteentitlement"]),
        ),
    )
    def test_valid_names(self, p_name, expected, base_entitlement_factory):
        entitlement = base_entitlement_factory(
            entitled=True,
            affordances={"presentedAs": p_name},
        )
        assert expected == entitlement.valid_names

    @pytest.mark.parametrize(
        "affordances,expected_value",
        (
            ({}, "testconcreteentitlement"),
            ({"presentedAs": "something_else"}, "something_else"),
        ),
    )
    def test_presentation_name(
        self, affordances, expected_value, entitlement_factory
    ):
        entitlement = entitlement_factory(
            ConcreteTestEntitlement,
            entitled=True,
            affordances=affordances,
        )
        assert expected_value == entitlement.presentation_name


class TestEntitlementCanEnable:
    def test_can_enable_false_on_unentitled(self, base_entitlement_factory):
        """When entitlement contract is not enabled, can_enable is False."""
        entitlement = base_entitlement_factory(
            entitled=False,
        )

        can_enable, reason = entitlement.can_enable()
        assert not can_enable
        assert reason.reason == CanEnableFailureReason.NOT_ENTITLED
        assert (
            reason.message.msg
            == messages.UNENTITLED.format(
                title=ConcreteTestEntitlement.title
            ).msg
        )

    def test_can_enable_false_on_access_only_not_supported(
        self, base_entitlement_factory
    ):
        """When entitlement contract is not enabled, can_enable is False."""
        entitlement = base_entitlement_factory(
            entitled=True,
            extra_args={
                "supports_access_only": False,
                "access_only": True,
            },
        )

        can_enable, reason = entitlement.can_enable()
        assert not can_enable
        assert (
            reason.reason == CanEnableFailureReason.ACCESS_ONLY_NOT_SUPPORTED
        )
        assert (
            reason.message.msg
            == messages.ENABLE_ACCESS_ONLY_NOT_SUPPORTED.format(
                title=ConcreteTestEntitlement.title
            ).msg
        )

    @pytest.mark.parametrize("caplog_text", [logging.DEBUG], indirect=True)
    @mock.patch("uaclient.contract.refresh")
    def test_can_enable_updates_expired_contract(
        self,
        m_refresh,
        caplog_text,
        base_entitlement_factory,
    ):
        """When entitlement contract is not enabled, can_enable is False."""
        entitlement = base_entitlement_factory(
            entitled=False,
        )

        with mock.patch.object(
            entitlement, "is_access_expired", return_value=True
        ):
            assert not entitlement.can_enable()[0]

        assert [mock.call(entitlement.cfg)] == m_refresh.call_args_list
        assert (
            "Updating contract on service 'testconcreteentitlement' expiry"
            in caplog_text()
        )

    def test_can_enable_false_on_entitlement_active(
        self, base_entitlement_factory
    ):
        """When entitlement is ENABLED, can_enable returns False."""
        entitlement = base_entitlement_factory(
            entitled=True,
            extra_args={
                "application_status": (ApplicationStatus.ENABLED, ""),
            },
        )

        can_enable, reason = entitlement.can_enable()
        assert not can_enable
        assert reason.reason == CanEnableFailureReason.ALREADY_ENABLED
        assert (
            reason.message.msg
            == messages.ALREADY_ENABLED.format(
                title=ConcreteTestEntitlement.title
            ).msg
        )

    @pytest.mark.parametrize(
        "applicability_status,expected_ret,expected_reason",
        (
            (
                (ApplicabilityStatus.INAPPLICABLE, "msg"),
                False,
                CanEnableFailure(CanEnableFailureReason.INAPPLICABLE, "msg"),
            ),
            (
                (ApplicabilityStatus.APPLICABLE, ""),
                True,
                None,
            ),
        ),
    )
    def test_can_enable_on_entitlement_given_applicability_status(
        self,
        applicability_status,
        expected_ret,
        expected_reason,
        base_entitlement_factory,
    ):
        """When entitlement is INAPPLICABLE, can_enable returns False."""
        entitlement = base_entitlement_factory(
            entitled=True,
            extra_args={
                "application_status": (ApplicationStatus.DISABLED, ""),
                "applicability_status": applicability_status,
            },
        )

        can_enable, reason = entitlement.can_enable()
        assert expected_ret is can_enable

        if reason:
            assert expected_reason.reason == reason.reason
            assert expected_reason.message == reason.message
        else:
            assert expected_reason is None

    def test_can_enable_when_incompatible_service_found(
        self, base_entitlement_factory, mock_entitlement
    ):
        entitlement = base_entitlement_factory(
            entitled=True,
            extra_args={
                "applicability_status": (ApplicabilityStatus.APPLICABLE, ""),
                "application_status": (ApplicationStatus.DISABLED, ""),
            },
        )
        m_incompatible_cls, _ = mock_entitlement(
            application_status=(ApplicationStatus.ENABLED, "")
        )

        entitlement._incompatible_services = (
            base.EntitlementWithMessage(
                m_incompatible_cls, messages.NamedMessage("test", "test")
            ),
        )

        ret, reason = entitlement.can_enable()

        assert ret is False
        assert reason.reason == CanEnableFailureReason.INCOMPATIBLE_SERVICE
        assert reason.message is None

    def test_can_enable_when_required_service_found(
        self, base_entitlement_factory, mock_entitlement
    ):
        m_required_service_cls, _ = mock_entitlement(
            application_status=(ApplicationStatus.DISABLED, "")
        )

        entitlement = base_entitlement_factory(
            entitled=True,
            extra_args={
                "applicability_status": (ApplicabilityStatus.APPLICABLE, ""),
                "application_status": (ApplicationStatus.DISABLED, ""),
                "required_services": (
                    mock.MagicMock(entitlement=m_required_service_cls),
                ),
            },
        )

        ret, reason = entitlement.can_enable()

        assert ret is False
        assert (
            reason.reason == CanEnableFailureReason.INACTIVE_REQUIRED_SERVICES
        )
        assert reason.message is None


class TestEntitlementEnable:
    @pytest.mark.parametrize(
        "block_disable_on_enable",
        [(False, False), (False, True), (True, False), (True, True)],
    )
    @mock.patch("uaclient.util.is_config_value_true")
    def test_enable_when_incompatible_service_found(
        self,
        m_is_config_value_true,
        block_disable_on_enable,
        base_entitlement_factory,
        mock_entitlement,
    ):
        m_is_config_value_true.return_value = block_disable_on_enable
        entitlement = base_entitlement_factory(
            entitled=True,
            extra_args={
                "applicability_status": (ApplicabilityStatus.APPLICABLE, ""),
                "application_status": (ApplicationStatus.DISABLED, ""),
                "enable": True,
            },
        )

        m_incompatible_cls, m_incompatible_obj = mock_entitlement(
            application_status=(ApplicationStatus.ENABLED, ""),
            disable=(True, None),
        )
        entitlement._incompatible_services = (
            base.EntitlementWithMessage(
                m_incompatible_cls, messages.NamedMessage("test", "test")
            ),
        )

        ret, reason = entitlement.enable(mock.MagicMock())

        expected_ret = False
        expected_reason = CanEnableFailureReason.INCOMPATIBLE_SERVICE
        if not block_disable_on_enable:
            expected_ret = True
            expected_reason = None
        expected_disable_call = 1 if expected_ret else 0

        assert ret == expected_ret
        if expected_reason is None:
            assert reason is None
        else:
            assert reason.reason == expected_reason
        assert m_is_config_value_true.call_count == 1
        assert m_incompatible_obj.disable.call_count == expected_disable_call

    def test_enable_when_required_service_found(
        self, base_entitlement_factory, mock_entitlement
    ):
        m_required_service_cls, m_required_service_obj = mock_entitlement(
            application_status=(ApplicationStatus.DISABLED, ""),
            enable=(True, None),
        )

        entitlement = base_entitlement_factory(
            entitled=True,
            extra_args={
                "applicability_status": (ApplicabilityStatus.APPLICABLE, ""),
                "application_status": (ApplicationStatus.DISABLED, ""),
                "enable": True,
                "required_services": (
                    mock.MagicMock(entitlement=m_required_service_cls),
                ),
            },
        )

        ret, reason = entitlement.enable(mock.MagicMock())

        assert ret is True
        assert reason is None
        assert m_required_service_obj.enable.call_count == 1

    @pytest.mark.parametrize(
        "can_enable_fail,handle_incompat_calls,enable_req_calls",
        [
            (
                CanEnableFailure(
                    CanEnableFailureReason.NOT_ENTITLED, message="msg"
                ),
                0,
                0,
            ),
            (
                CanEnableFailure(
                    CanEnableFailureReason.ALREADY_ENABLED,
                    message="msg",
                ),
                0,
                0,
            ),
            (
                CanEnableFailure(CanEnableFailureReason.INAPPLICABLE, "msg"),
                0,
                0,
            ),
            (
                CanEnableFailure(CanEnableFailureReason.INCOMPATIBLE_SERVICE),
                1,
                0,
            ),
            (
                CanEnableFailure(
                    CanEnableFailureReason.INACTIVE_REQUIRED_SERVICES
                ),
                0,
                1,
            ),
        ],
    )
    @mock.patch(
        "uaclient.entitlements.base.UAEntitlement._enable_required_services",
        return_value=(False, "required error msg"),
    )
    @mock.patch(
        "uaclient.entitlements.base.UAEntitlement.handle_incompatible_services",  # noqa: E501
        return_value=False,
    )
    @mock.patch("uaclient.entitlements.base.UAEntitlement.can_enable")
    def test_enable_false_when_can_enable_false(
        self,
        m_can_enable,
        m_handle_incompat,
        m_enable_required,
        can_enable_fail,
        handle_incompat_calls,
        enable_req_calls,
        base_entitlement_factory,
    ):
        """When can_enable returns False enable returns False."""
        m_can_enable.return_value = (False, can_enable_fail)
        m_handle_incompat.return_value = (False, None)
        entitlement = base_entitlement_factory(entitled=True)
        entitlement._perform_enable = mock.Mock()

        assert (False, can_enable_fail) == entitlement.enable(mock.MagicMock())

        assert 1 == m_can_enable.call_count
        assert handle_incompat_calls == m_handle_incompat.call_count
        assert enable_req_calls == m_enable_required.call_count
        assert 0 == entitlement._perform_enable.call_count

        if enable_req_calls:
            assert can_enable_fail.message == "required error msg"

    @pytest.mark.parametrize("enable_fail_message", (("not entitled"), (None)))
    @mock.patch("uaclient.util.handle_message_operations")
    @mock.patch("uaclient.util.prompt_for_confirmation")
    def test_enable_false_when_fails_to_enable_required_service(
        self,
        m_handle_msg,
        m_prompt_for_confirmation,
        enable_fail_message,
        base_entitlement_factory,
        mock_entitlement,
    ):
        m_handle_msg.return_value = True

        fail_reason = CanEnableFailure(
            CanEnableFailureReason.INACTIVE_REQUIRED_SERVICES
        )

        if enable_fail_message:
            msg = messages.NamedMessage("test-code", enable_fail_message)
        else:
            msg = None

        enable_fail_reason = CanEnableFailure(
            CanEnableFailureReason.NOT_ENTITLED, message=msg
        )

        m_prompt_for_confirmation.return_vale = True

        m_required_service_cls, m_required_service_obj = mock_entitlement(
            application_status=(ApplicationStatus.DISABLED, ""),
            title="Test",
            enable=(False, enable_fail_reason),
        )

        entitlement = base_entitlement_factory(
            entitled=True,
            extra_args={
                "application_status": (ApplicationStatus.DISABLED, ""),
                "required_services": (
                    mock.MagicMock(entitlement=m_required_service_cls),
                ),
            },
        )

        with mock.patch.object(entitlement, "can_enable") as m_can_enable:
            m_can_enable.return_value = (False, fail_reason)
            ret, fail = entitlement.enable(mock.MagicMock())

        assert not ret
        expected_msg = "Cannot enable required service: Test"
        if enable_fail_reason.message:
            expected_msg += "\n" + enable_fail_reason.message.msg
        assert expected_msg == fail.message.msg
        assert 1 == m_can_enable.call_count

    @mock.patch("uaclient.util.handle_message_operations")
    def test_enable_fails_when_blocking_service_is_enabled(
        self,
        _m_handle_message_op,
        mock_entitlement,
        base_entitlement_factory,
    ):
        expected_msg = messages.E_INCOMPATIBLE_SERVICE_STOPS_ENABLE.format(
            service_being_enabled=ConcreteTestEntitlement.title,
            incompatible_service="Test",
        )
        incompatible_service_cls, _ = mock_entitlement(
            application_status=(ApplicationStatus.ENABLED, ""),
            title="Test",
            disable=(
                False,
                CanDisableFailure(
                    CanDisableFailureReason.ACTIVE_DEPENDENT_SERVICES,
                    message=expected_msg,
                ),
            ),
        )
        incompatible_services_definition = base.EntitlementWithMessage(
            incompatible_service_cls, messages.NamedMessage("code", "msg")
        )
        entitlement = base_entitlement_factory(
            extra_args={
                "blocking_incompatible_services": [
                    incompatible_services_definition
                ],
                "incompatible_services": [incompatible_services_definition],
            },
        )

        result, reason = entitlement.enable(mock.MagicMock())
        assert not result
        assert expected_msg.msg == reason.message.msg.strip()


class TestEntitlementCanDisable:
    def test_can_disable_false_on_purge_not_supported(
        self, base_entitlement_factory
    ):
        """When the entitlement doesn't support purge, can_disable is FALSE."""
        entitlement = base_entitlement_factory(
            entitled=True,
            purge=True,
            extra_args={
                "supports_purge": False,
                "application_status": (ApplicationStatus.ENABLED, ""),
            },
        )

        can_disable, reason = entitlement.can_disable()
        assert can_disable is False
        assert reason.reason == CanDisableFailureReason.PURGE_NOT_SUPPORTED
        assert (
            reason.message.msg
            == messages.DISABLE_PURGE_NOT_SUPPORTED.format(
                title=ConcreteTestEntitlement.title
            ).msg
        )

    @pytest.mark.parametrize(
        "application_status,expected_ret,expected_fail",
        (
            (
                (ApplicationStatus.DISABLED, ""),
                False,
                CanDisableFailure(
                    CanDisableFailureReason.ALREADY_DISABLED,
                    messages.ALREADY_DISABLED.format(
                        title=ConcreteTestEntitlement.title
                    ),
                ),
            ),
            (
                (ApplicationStatus.ENABLED, ""),
                True,
                None,
            ),
        ),
    )
    def test_can_disable_given_entitlement_status(
        self,
        application_status,
        expected_ret,
        expected_fail,
        base_entitlement_factory,
    ):
        entitlement = base_entitlement_factory(
            entitled=True,
            extra_args={
                "application_status": application_status,
            },
        )

        ret, fail = entitlement.can_disable()
        assert expected_ret is ret

        if fail:
            assert expected_fail.message.msg == fail.message.msg
            assert expected_fail.reason == fail.reason
        else:
            assert expected_fail is None

    def test_can_disable_false_on_dependent_service(
        self, base_entitlement_factory, mock_entitlement
    ):
        m_required_service_cls, _ = mock_entitlement(
            application_status=(ApplicationStatus.ENABLED, None)
        )

        entitlement = base_entitlement_factory(
            entitled=True,
            extra_args={
                "application_status": (ApplicationStatus.ENABLED, ""),
                "dependent_services": (m_required_service_cls,),
            },
        )

        ret, fail = entitlement.can_disable()
        assert not ret
        assert fail.reason == CanDisableFailureReason.ACTIVE_DEPENDENT_SERVICES
        assert fail.message is None

    @mock.patch("uaclient.entitlements.entitlement_factory")
    def test_can_disable_when_ignoring_dependent_service(
        self, m_ent_factory, base_entitlement_factory, mock_entitlement
    ):
        entitlement = base_entitlement_factory(
            entitled=True,
            extra_args={
                "application_status": (ApplicationStatus.ENABLED, ""),
                "dependent_services": ("test",),
            },
        )

        m_dependent_service_cls, _ = mock_entitlement(
            application_status=(ApplicationStatus.ENABLED, None)
        )
        m_ent_factory.return_value = m_dependent_service_cls

        ret, fail = entitlement.can_disable(ignore_dependent_services=True)
        assert ret is True
        assert fail is None


class TestEntitlementDisable:
    def test_disable_when_dependent_service_found(
        self, base_entitlement_factory, mock_entitlement
    ):
        m_dependent_service_cls, m_dependent_service_obj = mock_entitlement(
            application_status=(ApplicationStatus.ENABLED, ""),
            disable=(True, None),
        )
        entitlement = base_entitlement_factory(
            entitled=True,
            extra_args={
                "application_status": (ApplicationStatus.ENABLED, ""),
                "disable": True,
                "dependent_services": (m_dependent_service_cls,),
            },
        )

        ret, fail = entitlement.disable(mock.MagicMock())

        expected_ret = True
        expected_disable_call = 1

        assert ret == expected_ret
        assert fail is None
        assert (
            m_dependent_service_obj.disable.call_count == expected_disable_call
        )

    @pytest.mark.parametrize("disable_fail_message", (("error"), (None)))
    @mock.patch("uaclient.util.handle_message_operations")
    def test_disable_false_when_fails_to_disable_dependent_service(
        self,
        m_handle_msg,
        disable_fail_message,
        base_entitlement_factory,
        mock_entitlement,
    ):
        m_handle_msg.return_value = True

        fail_reason = CanDisableFailure(
            CanDisableFailureReason.ACTIVE_DEPENDENT_SERVICES
        )

        msg = None
        if disable_fail_message:
            msg = messages.NamedMessage("test-code", disable_fail_message)

        disable_fail_reason = CanDisableFailure(
            CanDisableFailureReason.ALREADY_DISABLED, message=msg
        )
        m_dependent_service_cls, _ = mock_entitlement(
            name="Test",
            title="Test",
            disable=(False, disable_fail_reason),
            application_status=(ApplicationStatus.ENABLED, ""),
        )

        entitlement = base_entitlement_factory(
            entitled=True,
            extra_args={
                "application_status": (ApplicationStatus.DISABLED, ""),
                "dependent_services": (m_dependent_service_cls,),
            },
        )

        with mock.patch.object(entitlement, "can_disable") as m_can_disable:
            m_can_disable.return_value = (False, fail_reason)
            ret, fail = entitlement.disable(mock.MagicMock())

        assert not ret
        expected_msg = messages.FAILED_DISABLING_DEPENDENT_SERVICE.format(
            required_service="Test",
            error="\n" + msg.msg if msg else "",
        ).msg
        assert expected_msg == fail.message.msg
        assert 1 == m_can_disable.call_count

    def test_disable_returns_false_on_can_disable_false_and_does_nothing(
        self,
        base_entitlement_factory,
    ):
        """When can_disable is false disable returns false and noops."""
        entitlement = base_entitlement_factory()

        with mock.patch.object(
            entitlement, "can_disable", return_value=(False, None)
        ) as m_can_disable:
            ret, fail = entitlement.disable(mock.MagicMock())

        assert ret is False
        assert fail is None
        assert [mock.call()] == m_can_disable.call_args_list


class TestEntitlementContractStatus:
    @pytest.mark.parametrize(
        "entitled,expected_contract_status",
        (
            (
                True,
                ContractStatus.ENTITLED,
            ),
            (
                False,
                ContractStatus.UNENTITLED,
            ),
        ),
    )
    def test_contract_status(
        self, entitled, expected_contract_status, base_entitlement_factory
    ):
        entitlement = base_entitlement_factory(entitled=entitled)
        assert expected_contract_status == entitlement.contract_status()


class TestUserFacingStatus:
    @pytest.mark.parametrize(
        "entitled,applicability_status,application_status,expected_status,expected_details",  # noqa
        (
            (
                True,
                (
                    ApplicabilityStatus.INAPPLICABLE,
                    None,
                ),
                (
                    ApplicationStatus.DISABLED,
                    None,
                ),
                UserFacingStatus.INAPPLICABLE,
                None,
            ),
            (
                False,
                (
                    ApplicabilityStatus.APPLICABLE,
                    "",
                ),
                (
                    ApplicationStatus.DISABLED,
                    None,
                ),
                UserFacingStatus.UNAVAILABLE,
                messages.SERVICE_NOT_ENTITLED.format(
                    title=ConcreteTestEntitlement.title
                ),
            ),
            (
                True,
                (
                    ApplicabilityStatus.APPLICABLE,
                    "",
                ),
                (
                    ApplicationStatus.DISABLED,
                    None,
                ),
                UserFacingStatus.INACTIVE,
                None,
            ),
            (
                True,
                (
                    ApplicabilityStatus.APPLICABLE,
                    "",
                ),
                (
                    ApplicationStatus.WARNING,
                    None,
                ),
                UserFacingStatus.WARNING,
                None,
            ),
        ),
    )
    def test_user_facing_status(
        self,
        entitled,
        applicability_status,
        application_status,
        expected_status,
        expected_details,
        base_entitlement_factory,
    ):
        entitlement = base_entitlement_factory(
            entitled=entitled,
            extra_args={
                "applicability_status": applicability_status,
                "application_status": application_status,
            },
        )

        user_facing_status, details = entitlement.user_facing_status()
        assert expected_status == user_facing_status

        if details:
            assert expected_details.msg == details.msg
        else:
            assert expected_details is None

    def test_unavailable_when_applicable_but_no_entitlement_cfg(
        self, base_entitlement_factory
    ):
        entitlement = base_entitlement_factory(
            entitled=False,
            extra_args={
                "applicability_status": (ApplicabilityStatus.APPLICABLE, "")
            },
        )
        with mock.patch.object(
            entitlement, "_base_entitlement_cfg", return_value={}
        ):
            user_facing_status, details = entitlement.user_facing_status()

        assert UserFacingStatus.UNAVAILABLE == user_facing_status
        expected_details = messages.SERVICE_NOT_ENTITLED.format(
            title=entitlement.title
        ).msg
        assert expected_details == details.msg


class TestApplicabilityStatus:
    @pytest.mark.parametrize(
        (
            "arch,series,version,min_kernel,kernel_flavors,"
            "expected_status,expected_message"
        ),
        (
            (
                "arm64",
                "xenial",
                "16.04 LTS (Xenial Xerus)",
                "4.4",
                ["generic", "lowlatency"],
                ApplicabilityStatus.INAPPLICABLE,
                messages.INAPPLICABLE_ARCH.format(
                    title=ConcreteTestEntitlement.title,
                    arch="arm64",
                    supported_arches=", ".join(["amd64", "s390x"]),
                ),
            ),
            (
                "s390x",
                "bionic",
                "18.04 LTS (Bionic Beaver)",
                "4.4",
                ["generic", "lowlatency"],
                ApplicabilityStatus.INAPPLICABLE,
                messages.INAPPLICABLE_SERIES.format(
                    title=ConcreteTestEntitlement.title,
                    series="18.04 LTS (Bionic Beaver)",
                ),
            ),
            (
                "s390x",
                "xenial",
                "16.04 LTS (Xenial Xerus)",
                "5.0",
                ["generic", "lowlatency"],
                ApplicabilityStatus.INAPPLICABLE,
                messages.INAPPLICABLE_KERNEL_VER.format(
                    title=ConcreteTestEntitlement.title,
                    kernel="4.19.0-00-generic",
                    min_kernel="5.0",
                ),
            ),
            (
                "s390x",
                "xenial",
                "16.04 LTS (Xenial Xerus)",
                "5.0",
                ["lowlatency"],
                ApplicabilityStatus.INAPPLICABLE,
                messages.INAPPLICABLE_KERNEL.format(
                    title=ConcreteTestEntitlement.title,
                    kernel="4.19.0-00-generic",
                    supported_kernels="lowlatency",
                ),
            ),
            (
                "s390x",
                "xenial",
                "16.04 LTS (Xenial Xerus)",
                "4.4",
                ["generic", "lowlatency"],
                ApplicabilityStatus.APPLICABLE,
                None,
            ),
        ),
    )
    @mock.patch("uaclient.system.get_kernel_info")
    @mock.patch("uaclient.system.get_dpkg_arch")
    @mock.patch("uaclient.system.get_release_info")
    def test_applicability_status(
        self,
        m_release_info,
        m_dpkg_arch,
        m_get_kernel_info,
        arch,
        series,
        version,
        min_kernel,
        kernel_flavors,
        expected_status,
        expected_message,
        base_entitlement_factory,
    ):
        m_release_info.return_value = system.ReleaseInfo(
            distribution="", release="", series=series, pretty_version=version
        )
        m_get_kernel_info.return_value = system.KernelInfo(
            uname_machine_arch="",
            uname_release="4.19.0-00-generic",
            proc_version_signature_version="",
            build_date=None,
            major=4,
            minor=19,
            patch=0,
            abi="00",
            flavor="generic",
        )

        m_dpkg_arch.return_value = arch

        entitlement = base_entitlement_factory(
            directives={
                "aptURL": "http://CC",
                "aptKey": "APTKEY",
                "suites": ["xenial"],
                "additionalPackages": ["test-package"],
            },
            affordances={
                "architectures": ["x86_64", "s390x"],
                "series": ["xenial"],
                "minKernelVersion": min_kernel,
                "kernelFlavors": kernel_flavors,
            },
        )

        actual_status, actual_message = entitlement.applicability_status()
        assert expected_status == actual_status

        if expected_message:
            assert expected_message.msg == actual_message.msg
        else:
            assert actual_message is None


class TestEntitlementProcessContractDeltas:
    @pytest.mark.parametrize(
        "orig_access,delta",
        (({}, {}), ({}, {"entitlement": {"entitled": False}})),
    )
    def test_process_contract_deltas_does_nothing_on_empty_orig_access(
        self, base_entitlement_factory, orig_access, delta
    ):
        """When orig_acccess dict is empty perform no work."""
        entitlement = base_entitlement_factory(
            entitled=True,
            extra_args={
                "applicability_status": (ApplicabilityStatus.APPLICABLE, ""),
                "application_status": (ApplicationStatus.DISABLED, ""),
            },
        )
        with mock.patch.object(entitlement, "can_disable") as m_can_disable:
            entitlement.process_contract_deltas(orig_access, delta)
        assert 0 == m_can_disable.call_count

    @pytest.mark.parametrize(
        "orig_access,delta",
        (
            ({"entitlement": {"entitled": True}}, {}),  # no deltas
            (
                {"entitlement": {"entitled": False}},
                {"entitlement": {"entitled": True}},
            ),  # transition to entitled
            (
                {"entitlement": {"entitled": False}},
                {
                    "entitlement": {
                        "entitled": False,  # overridden by series 'example'
                        "series": {"example": {"entitled": True}},
                    }
                },
            ),
        ),
    )
    @mock.patch(
        "uaclient.system.get_release_info",
        return_value=system.ReleaseInfo(
            distribution="", release="", series="example", pretty_version=""
        ),
    )
    def test_process_contract_deltas_does_nothing_when_delta_remains_entitled(
        self, m_platform_info, base_entitlement_factory, orig_access, delta
    ):
        """If deltas do not represent transition to unentitled, do nothing."""
        entitlement = base_entitlement_factory(
            entitled=True,
            extra_args={
                "applicability_status": (ApplicabilityStatus.APPLICABLE, ""),
                "application_status": (ApplicationStatus.DISABLED, ""),
            },
        )
        entitlement.process_contract_deltas(orig_access, delta)
        assert (
            ApplicationStatus.DISABLED,
            mock.ANY,
        ) == entitlement.application_status()

    @pytest.mark.parametrize(
        "orig_access,delta",
        (
            (
                {
                    "entitlement": {"entitled": True}
                },  # Full entitlement dropped
                {"entitlement": {"entitled": util.DROPPED_KEY}},
            ),
            (
                {"entitlement": {"entitled": True}},
                {"entitlement": {"entitled": False}},
            ),  # transition to unentitled
        ),
    )
    def test_process_contract_deltas_clean_cache_on_inactive_unentitled(
        self, base_entitlement_factory, orig_access, delta, caplog_text
    ):
        """Only clear cache when deltas transition inactive to unentitled."""
        entitlement = base_entitlement_factory(
            entitled=True,
            extra_args={
                "application_status": (ApplicationStatus.DISABLED, ""),
            },
        )
        entitlement.process_contract_deltas(orig_access, delta)
        # If an entitlement is disabled, we don't need to tell the user
        # anything about it becoming unentitled
        # (FIXME: Something on bionic means that DEBUG log lines are being
        # picked up by caplog_text(), so work around that here)
        assert [] == [
            line for line in caplog_text().splitlines() if "DEBUG" not in line
        ]

    @pytest.mark.parametrize(
        "orig_access,delta",
        (
            (
                {
                    "entitlement": {"entitled": True}
                },  # Full entitlement dropped
                {"entitlement": {"entitled": util.DROPPED_KEY}},
            ),
            (
                {"entitlement": {"entitled": True}},
                {"entitlement": {"entitled": False}},
            ),  # transition to unentitled
        ),
    )
    def test_process_contract_deltas_disable_on_active_unentitled(
        self, base_entitlement_factory, orig_access, delta
    ):
        """Disable when deltas transition from active to unentitled."""
        entitlement = base_entitlement_factory(
            entitled=True,
            extra_args={
                "application_status": (ApplicationStatus.ENABLED, ""),
            },
        )

        with mock.patch.object(
            entitlement, "can_disable", return_value=(True, None)
        ):
            entitlement.process_contract_deltas(orig_access, delta)

        assert (
            ApplicationStatus.DISABLED,
            mock.ANY,
        ) == entitlement.application_status()


class TestEntitlementCfg:
    @pytest.mark.parametrize(
        "variant_name", ((""), ("test-variant"), ("invalid-variant"))
    )
    def test_entitlement_cfg_respects_variant(
        self, variant_name, base_entitlement_factory
    ):
        entitlement = base_entitlement_factory(
            entitled=True,
            obligations={"enableByDefault": False},
            affordances={
                "architectures": [
                    "amd64",
                    "ppc64el",
                ],
                "series": ["xenial", "bionic", "focal"],
            },
            directives={
                "additionalPackages": ["test-package"],
                "suites": ["xenial", "bionic", "focal"],
            },
            overrides=[
                {
                    "directives": {
                        "additionalPackages": ["test-package-variant"]
                    },
                    "selector": {
                        "variant": "test-variant",
                    },
                },
                {
                    "directives": {
                        "additionalPackages": ["test-package-unused"]
                    },
                    "selector": {"cloud": "aws", "series": "focal"},
                },
            ],
            extra_args={
                "applicability_status": (ApplicabilityStatus.APPLICABLE, ""),
                "application_status": (ApplicationStatus.DISABLED, ""),
                "variant_name": variant_name,
            },
        )
        expected_entitlement = copy.deepcopy(
            entitlement._base_entitlement_cfg()
        )
        if variant_name == "test-variant":
            expected_entitlement["entitlement"]["directives"][
                "additionalPackages"
            ] = ["test-package-variant"]

        assert expected_entitlement == entitlement.entitlement_cfg


class TestVariant:
    @pytest.mark.parametrize(
        "contract_variants",
        (
            ([]),
            (["not-found-variant"]),
            (["test_variant"]),
            (["test_variant", "test_variant2"]),
        ),
    )
    @mock.patch("uaclient.entitlements.base.UAEntitlement._get_variants")
    @mock.patch(
        "uaclient.entitlements.base.UAEntitlement._get_contract_variants"
    )
    def test_variant_property(
        self,
        m_get_contract_variants,
        m_get_variants,
        contract_variants,
        base_entitlement_factory,
    ):
        entitlement = base_entitlement_factory()
        service_variants = {"test_variant": "test", "generic": "generic"}
        m_get_contract_variants.return_value = contract_variants
        m_get_variants.return_value = service_variants
        actual_variants = entitlement.variants

        expected_variants = (
            {} if "test_variant" not in contract_variants else service_variants
        )
        assert expected_variants == actual_variants
        assert 1 == m_get_contract_variants.call_count
        assert 1 == m_get_variants.call_count


class TestGetContractVariant:
    @pytest.mark.parametrize(
        "overrides",
        (
            ([]),
            (
                [
                    {
                        "selector": {
                            "variant": "test1",
                        },
                    },
                    {
                        "selector": {
                            "variant": "test2",
                        }
                    },
                    {
                        "selector": {
                            "cloud": "cloud",
                        },
                    },
                ]
            ),
        ),
    )
    def test_get_contract_variant(self, overrides, base_entitlement_factory):
        entitlement = base_entitlement_factory(overrides=overrides)
        actual_contract_variants = entitlement._get_contract_variants()
        expected_contract_variants = (
            set() if not overrides else set(["test1", "test2"])
        )

        assert expected_contract_variants == actual_contract_variants


class TestHandleRequiredSnaps:
    @pytest.mark.parametrize(
        "directives",
        (
            ({}),
            (
                {
                    "requiredSnaps": [
                        {"name": "test1", "channel": "latest/stable"},
                        {
                            "name": "test2",
                            "classicConfinementSupport": True,
                        },
                        {"name": "test3"},
                    ]
                }
            ),
        ),
    )
    @mock.patch(
        "uaclient.snap.is_snapd_installed_as_a_snap", return_value=True
    )
    @mock.patch("uaclient.snap.is_snapd_installed", return_value=True)
    @mock.patch("uaclient.snap.run_snapd_wait_cmd")
    @mock.patch("uaclient.snap.get_snap_info")
    @mock.patch("uaclient.snap.install_snap")
    @mock.patch("uaclient.http.validate_proxy")
    @mock.patch("uaclient.snap.configure_snap_proxy")
    def test_handle_required_snaps(
        self,
        m_configure_snap_proxy,
        m_validate_proxy,
        m_install_snap,
        m_get_snap_info,
        m_run_snapd_wait_cmd,
        m_is_snapd_installed,
        m_is_snapd_installed_as_a_snap,
        directives,
        base_entitlement_factory,
    ):
        entitlement = base_entitlement_factory(directives=directives)
        m_get_snap_info.side_effect = [
            exceptions.SnapNotInstalledError(snap="snap"),
            exceptions.SnapNotInstalledError(snap="snap"),
            mock.MagicMock,
        ]

        assert entitlement.handle_required_snaps(mock.MagicMock())

        if not directives:
            assert 0 == m_is_snapd_installed.call_count
            assert 0 == m_is_snapd_installed_as_a_snap.call_count
            assert 0 == m_run_snapd_wait_cmd.call_count
            assert 0 == m_validate_proxy.call_count
            assert 0 == m_configure_snap_proxy.call_count
        else:
            assert 1 == m_is_snapd_installed.call_count
            assert 1 == m_is_snapd_installed_as_a_snap.call_count
            assert 1 == m_run_snapd_wait_cmd.call_count
            assert 2 == m_validate_proxy.call_count
            assert 1 == m_configure_snap_proxy.call_count
            assert [
                mock.call(
                    "test1",
                    channel="latest/stable",
                    classic_confinement_support=False,
                ),
                mock.call(
                    "test2", channel=None, classic_confinement_support=True
                ),
            ] == m_install_snap.call_args_list


class TestHandleRequiredPackages:
    @pytest.mark.parametrize(
        [
            "required_packages_directive",
            "expected_apt_update_calls",
            "expected_apt_install_calls",
            "expected_result",
        ],
        [
            (
                None,
                [],
                [],
                True,
            ),
            (
                [],
                [],
                [],
                True,
            ),
            (
                [{"name": "package"}],
                [mock.call("/etc/apt/sources.list")],
                [mock.call(["package"])],
                True,
            ),
            (
                [{"name": "package"}, {"name": "package2"}],
                [mock.call("/etc/apt/sources.list")],
                [mock.call(["package", "package2"])],
                True,
            ),
        ],
    )
    @mock.patch(
        "uaclient.apt.get_system_sources_file",
        return_value="/etc/apt/sources.list",
    )
    @mock.patch("uaclient.apt.update_sources_list")
    @mock.patch("uaclient.apt.run_apt_install_command")
    @mock.patch("uaclient.apt.run_apt_update_command")
    def test_handle_required_packages(
        self,
        m_apt_update,
        m_apt_install,
        m_update_sources_list,
        m_get_system_sources_file,
        required_packages_directive,
        expected_apt_update_calls,
        expected_apt_install_calls,
        expected_result,
        base_entitlement_factory,
    ):
        entitlement = base_entitlement_factory(
            directives={"requiredPackages": required_packages_directive},
        )

        assert expected_result == entitlement.handle_required_packages(
            mock.MagicMock()
        )
        assert [] == m_apt_update.call_args_list
        assert (
            expected_apt_update_calls == m_update_sources_list.call_args_list
        )
        assert expected_apt_install_calls == m_apt_install.call_args_list

    @pytest.mark.parametrize(
        [
            "required_packages_directive",
            "expected_apt_remove_calls",
            "expected_result",
        ],
        [
            (
                None,
                [],
                True,
            ),
            (
                [],
                [],
                True,
            ),
            (
                [{"name": "package"}],
                [],
                True,
            ),
            (
                [{"name": "package"}, {"name": "package2"}],
                [],
                True,
            ),
            (
                [
                    {"name": "package"},
                    {"name": "package2", "removeOnDisable": True},
                ],
                [mock.call(["package2"], mock.ANY)],
                True,
            ),
            (
                [
                    {"name": "package"},
                    {"name": "package2", "removeOnDisable": True},
                    {"name": "package3", "removeOnDisable": False},
                    {"name": "package4", "removeOnDisable": True},
                ],
                [mock.call(["package2", "package4"], mock.ANY)],
                True,
            ),
        ],
    )
    @mock.patch("uaclient.apt.remove_packages")
    def test_handle_removing_required_packages(
        self,
        m_apt_remove,
        required_packages_directive,
        expected_apt_remove_calls,
        expected_result,
        base_entitlement_factory,
    ):
        entitlement = base_entitlement_factory(
            directives={"requiredPackages": required_packages_directive},
        )

        assert (
            expected_result == entitlement.handle_removing_required_packages()
        )
        assert expected_apt_remove_calls == m_apt_remove.call_args_list



import logging
import shlex
import subprocess
import time

from behave import step, then, when
from hamcrest import assert_that, equal_to

from features.util import SUT, process_template_vars


@when("I run `{command}` {user_spec}, retrying exit [{exit_codes}]")
def when_i_retry_run_command(
    context, command, user_spec, exit_codes, machine_name=SUT
):
    when_i_run_command(
        context,
        command,
        user_spec,
        verify_return=False,
        machine_name=machine_name,
    )
    retries = [5, 5, 10]  # Sleep times to wait between retries
    while len(retries) > 0 and str(
        context.process.returncode
    ) in exit_codes.split(","):
        time.sleep(retries.pop(0))
        logging.info(
            "--- Retrying on exit {exit_code}: {command}".format(
                exit_code=context.process.returncode, command=command
            )
        )
        when_i_run_command(
            context,
            command,
            user_spec,
            verify_return=False,
            machine_name=machine_name,
        )
    logging.warning(
        "Exhausted retries waiting for exit codes: %s. Final exit code: %d",
        exit_codes,
        context.process.returncode,
    )
    assert_that(context.process.returncode, equal_to(0))


@when("I run `{command}` `{user_spec}` and the following stdin")
def when_i_run_command_with_long_stdin(context, command, user_spec):
    text = process_template_vars(context, context.text)
    when_i_run_command(context, command, user_spec, stdin=text)


@when("I run `{command}` {user_spec}")
@when("I run `{command}` `{user_spec}` on the `{machine_name}` machine")
@when("I run `{command}` `{user_spec}` and stdin `{stdin}`")
def when_i_run_command(
    context,
    command,
    user_spec,
    verify_return=True,
    stdin=None,
    machine_name=SUT,
):
    command = process_template_vars(context, command)
    prefix = get_command_prefix_for_user_spec(user_spec)

    if context.pro_config.collect_coverage:
        coverage_command = (
            "python3 -m coverage run "
            "--source=/usr/lib/python3/dist-packages/uaclient /usr/bin/pro"
        )
        split_command = command.split()
        if split_command[0] == "pro":
            command = " ".join([coverage_command] + split_command[1:])

    full_cmd = prefix + shlex.split(command)

    if stdin is not None:
        stdin = stdin.replace("\\n", "\n")

    result = context.machines[machine_name].instance.execute(
        full_cmd, stdin=stdin
    )

    process = subprocess.CompletedProcess(
        args=full_cmd,
        stdout=result.stdout,
        stderr=result.stderr,
        returncode=result.return_code,
    )  # type: subprocess.CompletedProcess

    if verify_return and result.return_code != 0:
        logging.error("Error executing command: {}".format(command))
        logging.error("stdout: {}".format(result.stdout))
        logging.error("stderr: {}".format(result.stderr))
    else:
        logging.debug("stdout: {}".format(result.stdout))
        logging.debug("stderr: {}".format(result.stderr))

    if verify_return:
        assert_that(process.returncode, equal_to(0))

    context.process = process


@when("I run shell command `{command}` {user_spec}")
def when_i_run_shell_command(context, command, user_spec):
    when_i_run_command(context, 'sh -c "{}"'.format(command), user_spec)


@then("I verify that the `{cmd_name}` command is not found")
def then_i_should_see_that_the_command_is_not_found(context, cmd_name):
    cmd = "which {} || echo FAILURE".format(cmd_name)
    cmd = 'sh -c "{}"'.format(cmd)
    when_i_run_command(context, cmd, "as non-root")

    expected_return = "FAILURE"
    actual_return = context.process.stdout.strip()
    assert_that(expected_return, equal_to(actual_return))


@step(
    "I verify that running `{cmd_name}` `{spec}` and the following stdin exits `{exit_codes}`"  # noqa
)
def i_verify_that_running_cmd_with_spec_and_stdin_exits_with_codes(
    context, cmd_name, spec, exit_codes
):
    text = process_template_vars(context, context.text)
    then_i_verify_that_running_cmd_with_spec_and_stdin_exits_with_codes(
        context, cmd_name, spec, text, exit_codes
    )


@step(
    "I verify that running `{cmd_name}` `{spec}` and stdin `{stdin}` exits `{exit_codes}`"  # noqa
)
def then_i_verify_that_running_cmd_with_spec_and_stdin_exits_with_codes(
    context, cmd_name, spec, stdin, exit_codes
):
    when_i_run_command(
        context, cmd_name, spec, stdin=stdin, verify_return=False
    )

    expected_codes = exit_codes.split(",")
    if str(context.process.returncode) not in expected_codes:
        raise AssertionError(
            "Expected exit code in: {} but got {}".format(
                expected_codes, context.process.returncode
            )
        )


@step("I verify that running `{cmd_name}` `{spec}` exits `{exit_codes}`")
def then_i_verify_that_running_cmd_with_spec_exits_with_codes(
    context, cmd_name, spec, exit_codes
):
    when_i_run_command(context, cmd_name, spec, verify_return=False)
    logging.debug("got return code: %d", context.process.returncode)
    expected_codes = exit_codes.split(",")
    if str(context.process.returncode) not in expected_codes:
        raise AssertionError(
            "Expected exit code in: {} but got {}".format(
                expected_codes, context.process.returncode
            )
        )


def get_command_prefix_for_user_spec(user_spec):
    prefix = []
    if user_spec.lstrip() == "with sudo":
        prefix = ["sudo"]
    elif user_spec.lstrip() != "as non-root":
        raise Exception(
            "The two acceptable values for user_spec are: 'with sudo',"
            " 'as non-root'"
        )
    return prefix



import logging
from typing import Any, Dict, Optional, Tuple

from uaclient import api, event_logger, exceptions, messages, system, util
from uaclient.entitlements.base import UAEntitlement
from uaclient.entitlements.entitlement_status import (
    ApplicabilityStatus,
    ApplicationStatus,
)

LOG = logging.getLogger(util.replace_top_level_logger_name(__name__))
event = event_logger.get_event_logger()


class LandscapeEntitlement(UAEntitlement):
    name = "landscape"
    title = messages.LANDSCAPE_TITLE
    description = messages.LANDSCAPE_DESCRIPTION
    help_doc_url = messages.urls.LANDSCAPE_HOME_PAGE
    help_text = messages.LANDSCAPE_HELP_TEXT

    def enable_steps(self) -> int:
        return 1

    def disable_steps(self) -> int:
        return 1

    def _perform_enable(self, progress: api.ProgressWrapper) -> bool:
        cmd = ["landscape-config"] + self.extra_args
        if not progress.is_interactive() and "--silent" not in cmd:
            cmd += ["--silent"]

        LOG.debug("Executing: %r", cmd)
        progress.progress(
            util.redact_sensitive_logs(
                messages.EXECUTING_COMMAND.format(command=" ".join(cmd))
            )
        )
        try:
            system.subp(cmd, pipe_stdouterr=not progress.is_interactive())
        except exceptions.ProcessExecutionError as e:
            LOG.exception(e)
            if not progress.is_interactive():
                progress.emit("info", e.stderr.strip())
                raise exceptions.LandscapeConfigFailed(
                    stdout=e.stdout.strip(), stderr=e.stderr.strip()
                )
            return False
        return True

    def _perform_disable(self, progress: api.ProgressWrapper) -> bool:
        cmd = ["landscape-config", "--disable"]
        progress.progress(
            messages.EXECUTING_COMMAND.format(command=" ".join(cmd))
        )
        try:
            system.subp(cmd)
        except exceptions.ProcessExecutionError as e:
            LOG.error(e)
            progress.emit("info", str(e).strip())

        progress.emit("info", messages.LANDSCAPE_CONFIG_REMAINS)

        return True

    def application_status(
        self,
    ) -> Tuple[ApplicationStatus, Optional[messages.NamedMessage]]:
        if (
            self.are_required_packages_installed()
            and system.is_systemd_unit_active("landscape-client")
        ):
            return (ApplicationStatus.ENABLED, None)
        else:
            return (
                ApplicationStatus.DISABLED,
                messages.LANDSCAPE_SERVICE_NOT_ACTIVE,
            )

    def applicability_status(
        self,
    ) -> Tuple[ApplicabilityStatus, Optional[messages.NamedMessage]]:
        applicability_status = super().applicability_status()
        if applicability_status[0] == ApplicabilityStatus.INAPPLICABLE:
            affordance = self.entitlement_cfg["entitlement"].get(
                "affordances", {}
            )
            affordance_series = affordance.get("series", None)
            current_series = system.get_release_info().series
            if (
                self.affordance_check_series
                and affordance_series is not None
                and current_series not in affordance_series
            ):
                return (
                    ApplicabilityStatus.INAPPLICABLE,
                    messages.LANDSCAPE_INAPPLICABLE,
                )
        return applicability_status

    def enabled_warning_status(
        self,
    ) -> Tuple[bool, Optional[messages.NamedMessage]]:
        # This check wrongly gives warning when non-root
        # This will become obsolete soon: #2864
        if util.we_are_currently_root():
            try:
                system.subp(
                    ["landscape-config", "--is-registered", "--silent"]
                )
            except exceptions.ProcessExecutionError:
                return (
                    True,
                    messages.LANDSCAPE_NOT_REGISTERED,
                )

        return False, None

    def process_contract_deltas(
        self,
        orig_access: Dict[str, Any],
        deltas: Dict[str, Any],
        allow_enable: bool = False,
        verbose: bool = True,
    ) -> bool:
        # overriding allow_enable to always be False for this entitlement
        # effectively prevents enableByDefault from ever happening
        return super().process_contract_deltas(
            orig_access, deltas, allow_enable=False, verbose=verbose
        )



"""Tests related to uaclient.entitlement.base module."""

import copy
import logging
import os
from functools import partial

import mock
import pytest

import uaclient.entitlements.fips as fips
from uaclient import apt, exceptions, messages, system, util
from uaclient.clouds.identity import NoCloudTypeReason
from uaclient.entitlements.entitlement_status import (
    ApplicabilityStatus,
    ApplicationStatus,
    CanEnableFailureReason,
)
from uaclient.entitlements.fips import (
    CONDITIONAL_PACKAGES_EVERYWHERE,
    CONDITIONAL_PACKAGES_OPENSSH_HMAC,
    UBUNTU_FIPS_METAPACKAGE_DEPENDS_BIONIC,
    UBUNTU_FIPS_METAPACKAGE_DEPENDS_FOCAL,
    UBUNTU_FIPS_METAPACKAGE_DEPENDS_XENIAL,
    FIPSCommonEntitlement,
    FIPSEntitlement,
    FIPSPreviewEntitlement,
    FIPSUpdatesEntitlement,
)
from uaclient.files.notices import Notice, NoticesManager
from uaclient.testing import helpers

M_PATH = "uaclient.entitlements.fips."
M_LIVEPATCH_PATH = "uaclient.entitlements.livepatch.LivepatchEntitlement."
M_REPOPATH = "uaclient.entitlements.repo."
FIPS_ADDITIONAL_PACKAGES = ["ubuntu-fips"]


class ConcreteFIPSCommonEntitlement(FIPSCommonEntitlement):
    title = "TEST COMMON FIPS ENTITLEMENT"
    description = "TEST COMMON FIPS ENTITLEMENT"

    @property
    def conditional_packages(self):
        return self._conditional_packages

    @conditional_packages.setter
    def conditional_packages(self, value):
        self._conditional_packages = value

    @conditional_packages.deleter
    def conditional_packages(self):
        pass


@pytest.fixture(params=[FIPSEntitlement, FIPSUpdatesEntitlement])
def fips_entitlement_factory(request, entitlement_factory):
    """Parameterized fixture so we apply all tests to both FIPS and Updates"""
    additional_packages = FIPS_ADDITIONAL_PACKAGES
    return partial(
        entitlement_factory,
        request.param,
        additional_packages=additional_packages,
    )


@pytest.fixture
def entitlement(fips_entitlement_factory):
    return fips_entitlement_factory()


class TestFIPSEntitlementDefaults:
    @pytest.mark.parametrize(
        "series, is_container, expected",
        (
            (
                "xenial",
                False,
                CONDITIONAL_PACKAGES_EVERYWHERE
                + CONDITIONAL_PACKAGES_OPENSSH_HMAC,
            ),
            (
                "xenial",
                True,
                CONDITIONAL_PACKAGES_EVERYWHERE
                + CONDITIONAL_PACKAGES_OPENSSH_HMAC
                + UBUNTU_FIPS_METAPACKAGE_DEPENDS_XENIAL,
            ),
            (
                "bionic",
                False,
                CONDITIONAL_PACKAGES_EVERYWHERE
                + CONDITIONAL_PACKAGES_OPENSSH_HMAC,
            ),
            (
                "bionic",
                True,
                CONDITIONAL_PACKAGES_EVERYWHERE
                + CONDITIONAL_PACKAGES_OPENSSH_HMAC
                + UBUNTU_FIPS_METAPACKAGE_DEPENDS_BIONIC,
            ),
            ("focal", False, CONDITIONAL_PACKAGES_EVERYWHERE),
            (
                "focal",
                True,
                CONDITIONAL_PACKAGES_EVERYWHERE
                + UBUNTU_FIPS_METAPACKAGE_DEPENDS_FOCAL,
            ),
        ),
    )
    @mock.patch("uaclient.system.is_container")
    @mock.patch("uaclient.system.get_release_info")
    def test_conditional_packages(
        self,
        m_get_release_info,
        m_is_container,
        series,
        is_container,
        expected,
        entitlement,
    ):
        """Test conditional package respect series restrictions"""
        m_get_release_info.return_value = mock.MagicMock(series=series)
        m_is_container.return_value = is_container

        conditional_packages = entitlement.conditional_packages
        assert expected == conditional_packages

    @pytest.mark.parametrize(
        "fips_version, assume_yes, expected_continue",
        (
            (
                "0",
                True,
                True,
            ),
            (
                "0",
                False,
                False,
            ),
            (
                "999",
                True,
                True,
            ),
            (
                "999",
                False,
                True,
            ),
        ),
    )
    @mock.patch(M_PATH + "apt.get_pkg_candidate_version")
    @mock.patch(M_PATH + "util.prompt_for_confirmation")
    @mock.patch(M_PATH + "system.get_kernel_info")
    def test_kernel_downgrade(
        self,
        m_prompt_kernel_info,
        m_prompt_for_confirmation,
        m_pkg_candidate_version,
        fips_version,
        assume_yes,
        expected_continue,
        entitlement,
    ):
        """Test kernel downgrades block install if user denies prompt"""
        # if user is prompted for confirmation assume they say no
        if not assume_yes:
            m_prompt_for_confirmation.return_value = False
        else:
            m_prompt_for_confirmation.return_value = True
        m_prompt_kernel_info.return_value.proc_version_signature_version = (
            "6.8.0-22.22-generic"
        )
        m_pkg_candidate_version.return_value = fips_version
        install_continues = entitlement.prompt_if_kernel_downgrade(
            assume_yes=assume_yes
        )
        assert install_continues == expected_continue

    def test_default_repo_key_file(self, entitlement):
        """GPG keyring file is the same for both FIPS and FIPS with Updates"""
        assert entitlement.repo_key_file == "ubuntu-pro-fips.gpg"

    def test_default_repo_pinning(self, entitlement):
        """FIPS and FIPS with Updates repositories are pinned."""
        assert entitlement.repo_pin_priority == 1001

    @mock.patch("uaclient.system.is_container", return_value=True)
    def test_messaging_on_containers(
        self, _m_is_container, fips_entitlement_factory
    ):
        """FIPS and FIPS Updates have different messaging on containers"""
        entitlement = fips_entitlement_factory()

        expected_msging = {
            "fips": {
                "pre_enable": [
                    (
                        util.prompt_for_confirmation,
                        {
                            "msg": messages.PROMPT_FIPS_CONTAINER_PRE_ENABLE.format(  # noqa: E501
                                title="FIPS"
                            ),
                        },
                    )
                ],
                "pre_install": [
                    (
                        entitlement.prompt_if_kernel_downgrade,
                        {},
                    )
                ],
                "post_enable": [messages.FIPS_RUN_APT_UPGRADE],
                "pre_disable": [
                    (
                        util.prompt_for_confirmation,
                        {
                            "msg": messages.PROMPT_FIPS_PRE_DISABLE.format(
                                title="FIPS"
                            ),
                        },
                    )
                ],
            },
            "fips-updates": {
                "pre_enable": [
                    (
                        util.prompt_for_confirmation,
                        {
                            "msg": messages.PROMPT_FIPS_CONTAINER_PRE_ENABLE.format(  # noqa: E501
                                title="FIPS Updates"
                            ),
                        },
                    )
                ],
                "pre_install": [
                    (
                        entitlement.prompt_if_kernel_downgrade,
                        {},
                    )
                ],
                "post_enable": [messages.FIPS_RUN_APT_UPGRADE],
                "pre_disable": [
                    (
                        util.prompt_for_confirmation,
                        {
                            "msg": messages.PROMPT_FIPS_PRE_DISABLE.format(
                                title="FIPS Updates"
                            ),
                        },
                    )
                ],
            },
        }

        if entitlement.name in expected_msging:
            with mock.patch.object(
                entitlement, "auto_upgrade_all_on_enable", return_value=False
            ):
                assert (
                    expected_msging[entitlement.name] == entitlement.messaging
                )
        else:
            assert False, "Unknown entitlement {}".format(entitlement.name)

    @pytest.mark.parametrize(
        [
            "fips_packages",
            "current_kernel",
            "expected_pre_enable_messages",
        ],
        [
            ([], None, [(util.prompt_for_confirmation, {"msg": mock.ANY})]),
            (
                ["ubuntu-fips"],
                mock.MagicMock(flavor="generic"),
                [(util.prompt_for_confirmation, {"msg": mock.ANY})],
            ),
            (
                ["ubuntu-aws-fips"],
                mock.MagicMock(flavor="aws"),
                [(util.prompt_for_confirmation, {"msg": mock.ANY})],
            ),
            (
                ["ubuntu-azure-fips"],
                mock.MagicMock(flavor="azure"),
                [(util.prompt_for_confirmation, {"msg": mock.ANY})],
            ),
            (
                ["ubuntu-gcp-fips"],
                mock.MagicMock(flavor="gcp"),
                [(util.prompt_for_confirmation, {"msg": mock.ANY})],
            ),
            (
                ["ubuntu-fips"],
                mock.MagicMock(flavor="nonstandard"),
                [
                    (util.prompt_for_confirmation, {"msg": mock.ANY}),
                    (
                        util.prompt_for_confirmation,
                        {
                            "msg": messages.KERNEL_FLAVOR_CHANGE_WARNING_PROMPT.format(  # noqa: E501
                                variant="generic",
                                service="fips",
                                base_flavor="generic",
                                current_flavor="nonstandard",
                            )
                        },
                    ),
                ],
            ),
        ],
    )
    @mock.patch(
        "uaclient.entitlements.fips.FIPSCommonEntitlement.packages",
        new_callable=mock.PropertyMock,
    )
    @mock.patch("uaclient.system.get_kernel_info")
    @mock.patch("uaclient.system.is_container", return_value=False)
    def test_messaging_on_wrong_kernel(
        self,
        _m_is_container,
        m_get_kernel_info,
        m_fips_packages,
        fips_packages,
        current_kernel,
        expected_pre_enable_messages,
        FakeConfig,
    ):
        m_get_kernel_info.return_value = current_kernel
        m_fips_packages.return_value = fips_packages
        assert expected_pre_enable_messages == FIPSEntitlement(
            FakeConfig()
        ).messaging.get("pre_enable")


class TestFIPSEntitlementEnable:
    @pytest.mark.parametrize(
        "fips_common_enable_return_value, expected_remove_notice_calls",
        [
            (True, [mock.call(Notice.FIPS_INSTALL_OUT_OF_DATE)]),
            (False, []),
        ],
    )
    @mock.patch(
        "uaclient.entitlements.fips.FIPSCommonEntitlement._perform_enable"
    )
    @mock.patch("uaclient.files.notices.NoticesManager.remove")
    def test_enable_removes_out_of_date_notice_on_success(
        self,
        m_remove_notice,
        m_fips_common_enable,
        fips_common_enable_return_value,
        expected_remove_notice_calls,
        entitlement_factory,
    ):
        m_fips_common_enable.return_value = fips_common_enable_return_value
        fips_entitlement = entitlement_factory(FIPSEntitlement)
        assert (
            fips_common_enable_return_value
            is fips_entitlement._perform_enable(mock.MagicMock())
        )
        assert expected_remove_notice_calls == m_remove_notice.call_args_list

    @pytest.mark.parametrize(
        "repo_enable_return_value, expected_remove_notice_calls",
        [
            (
                True,
                [
                    mock.call(
                        Notice.WRONG_FIPS_METAPACKAGE_ON_CLOUD,
                    ),
                    mock.call(
                        Notice.FIPS_REBOOT_REQUIRED,
                    ),
                ],
            ),
            (False, []),
        ],
    )
    @mock.patch("uaclient.entitlements.repo.RepoEntitlement._perform_enable")
    @mock.patch("uaclient.files.notices.NoticesManager.remove")
    def test_enable_removes_wrong_met_notice_on_success(
        self,
        m_remove_notice,
        m_repo_enable,
        repo_enable_return_value,
        expected_remove_notice_calls,
        entitlement,
    ):
        m_repo_enable.return_value = repo_enable_return_value
        with mock.patch.object(fips, "services_once_enabled_file"):
            assert repo_enable_return_value is entitlement._perform_enable(
                mock.MagicMock()
            )
        assert (
            expected_remove_notice_calls == m_remove_notice.call_args_list[:2]
        )

    @mock.patch("uaclient.util.handle_message_operations")
    @mock.patch(
        M_LIVEPATCH_PATH + "application_status",
        return_value=((ApplicationStatus.DISABLED, "")),
    )
    @mock.patch("uaclient.system.is_container", return_value=False)
    def test_enable_fails_when_fips_update_service_is_enabled(
        self,
        m_is_container,
        m_livepatch,
        m_handle_message_op,
        entitlement_factory,
    ):
        m_handle_message_op.return_value = True
        fips_entitlement = entitlement_factory(FIPSEntitlement)
        base_path = "uaclient.entitlements.fips.FIPSUpdatesEntitlement"

        with mock.patch(
            "{}.application_status".format(base_path)
        ) as m_fips_update:
            with mock.patch.object(
                fips_entitlement, "_allow_fips_on_cloud_instance"
            ) as m_allow_fips_on_cloud:
                m_allow_fips_on_cloud.return_value = True
                m_fips_update.return_value = (
                    ApplicationStatus.ENABLED,
                    "",
                )
                result, reason = fips_entitlement.enable(mock.MagicMock())
                assert not result
                expected_msg = (
                    "Cannot enable FIPS when FIPS Updates is enabled."
                )
                assert expected_msg.strip() == reason.message.msg.strip()

    @mock.patch("uaclient.util.handle_message_operations")
    @mock.patch(
        M_LIVEPATCH_PATH + "application_status",
        return_value=((ApplicationStatus.DISABLED, "")),
    )
    @mock.patch("uaclient.system.is_container", return_value=False)
    def test_enable_fails_when_fips_updates_service_once_enabled(
        self,
        m_is_container,
        m_livepatch,
        m_handle_message_op,
        entitlement_factory,
    ):
        m_handle_message_op.return_value = True
        fake_dof = mock.MagicMock()
        fake_ua_file = mock.MagicMock()
        fake_ua_file.to_json.return_value = {"fips_updates": True}
        fake_dof.read.return_value = fake_ua_file
        fips_entitlement = entitlement_factory(FIPSEntitlement)

        with mock.patch.object(
            fips_entitlement, "_allow_fips_on_cloud_instance"
        ) as m_allow_fips_on_cloud:
            m_allow_fips_on_cloud.return_value = True
            with mock.patch.object(
                fips, "services_once_enabled_file", fake_dof
            ):
                result, reason = fips_entitlement.enable(mock.MagicMock())
                assert not result
                expected_msg = (
                    "Cannot enable FIPS because FIPS Updates was once enabled."
                )
                assert expected_msg.strip() == reason.message.msg.strip()

    @mock.patch("uaclient.system.get_release_info")
    @mock.patch("uaclient.entitlements.fips.get_cloud_type")
    @mock.patch("uaclient.util.handle_message_operations")
    @mock.patch("uaclient.system.is_container", return_value=False)
    def test_enable_fails_when_on_xenial_cloud_instance(
        self,
        m_is_container,
        m_handle_message_op,
        m_cloud_type,
        m_get_release_info,
        entitlement,
    ):
        m_handle_message_op.return_value = True
        m_cloud_type.return_value = ("gce", None)
        m_get_release_info.return_value = mock.MagicMock(series="xenial")
        base_path = "uaclient.entitlements.livepatch.LivepatchEntitlement"

        with mock.patch(
            "{}.application_status".format(base_path)
        ) as m_livepatch:
            m_livepatch.return_value = (ApplicationStatus.DISABLED, "")
            result, reason = entitlement.enable(mock.MagicMock())
            assert not result
            expected_msg = """\
            Ubuntu Xenial does not provide a GCP optimized FIPS kernel"""
            assert expected_msg.strip() in reason.message.msg.strip()

    @mock.patch("uaclient.system.get_release_info")
    @mock.patch("uaclient.util.is_config_value_true", return_value=False)
    @mock.patch("uaclient.entitlements.fips.get_cloud_type")
    @mock.patch("uaclient.util.handle_message_operations")
    @mock.patch("uaclient.system.is_container", return_value=False)
    def test_enable_fails_when_on_gcp_instance_with_default_fips(
        self,
        m_is_container,
        m_handle_message_op,
        m_get_cloud_type,
        m_is_config_value_true,
        m_get_release_info,
        entitlement,
    ):
        m_handle_message_op.return_value = True
        m_get_cloud_type.return_value = ("gce", None)
        m_get_release_info.return_value = mock.MagicMock(series="test")

        ent_name = entitlement.name
        fips_cls_name = "FIPS" if ent_name == "fips" else "FIPSUpdates"
        base_path = "uaclient.entitlements.fips.{}Entitlement".format(
            fips_cls_name
        )

        with mock.patch(
            "{}.application_status".format(base_path)
        ) as m_fips_status:
            m_fips_status.return_value = (
                ApplicationStatus.DISABLED,
                "",
            )
            result, reason = entitlement.enable(mock.MagicMock())
            assert not result
            expected_msg = """\
            Ubuntu Test does not provide a GCP optimized FIPS kernel"""
            assert expected_msg.strip() in reason.message.msg.strip()

    @pytest.mark.parametrize(
        "allow_default_fips_metapackage_on_gcp", ((True), (False))
    )
    @pytest.mark.parametrize("cloud_id", (("aws"), ("gce"), ("azure"), (None)))
    @pytest.mark.parametrize("series", (("xenial"), ("bionic")))
    @mock.patch("uaclient.util.is_config_value_true")
    def test_prevent_fips_on_xenial_or_focal_cloud_instances(
        self,
        m_is_config_value_true,
        series,
        cloud_id,
        allow_default_fips_metapackage_on_gcp,
        entitlement,
    ):
        def mock_config_value(config, path_to_value):
            if "allow_default_fips_metapackage_on_gcp" in path_to_value:
                return allow_default_fips_metapackage_on_gcp

            return False

        m_is_config_value_true.side_effect = mock_config_value
        actual_value = entitlement._allow_fips_on_cloud_instance(
            cloud_id=cloud_id, series=series
        )

        if cloud_id in ("azure", "aws") or cloud_id is None:
            assert actual_value
        elif all([cloud_id == "gce", allow_default_fips_metapackage_on_gcp]):
            assert actual_value
        elif all(
            [
                cloud_id == "gce",
                not allow_default_fips_metapackage_on_gcp,
                series == "xenial",
            ]
        ):
            assert not actual_value
        elif cloud_id == "gce":
            assert actual_value

    @pytest.mark.parametrize(
        "cfg_allow_default_fips_metapkg_on_gcp", ((True), (False))
    )
    @pytest.mark.parametrize(
        "additional_pkgs", (["ubuntu-fips"], ["ubuntu-gcp-fips", "test"])
    )
    @pytest.mark.parametrize("series", (("xenial"), ("bionic"), ("focal")))
    @mock.patch("uaclient.util.is_config_value_true")
    def test_prevent_default_fips_on_gcp_cloud(
        self,
        m_is_config_value,
        series,
        additional_pkgs,
        cfg_allow_default_fips_metapkg_on_gcp,
        fips_entitlement_factory,
    ):
        m_is_config_value.return_value = cfg_allow_default_fips_metapkg_on_gcp
        entitlement = fips_entitlement_factory(
            additional_packages=additional_pkgs
        )
        actual_value = entitlement._allow_fips_on_cloud_instance(
            cloud_id="gce", series=series
        )

        if all(
            [
                not cfg_allow_default_fips_metapkg_on_gcp,
                "ubuntu-gcp-fips" not in additional_pkgs,
                series not in ("bionic", "focal"),
            ]
        ):
            assert not actual_value
        else:
            assert actual_value

    @pytest.mark.parametrize("caplog_text", [logging.WARNING], indirect=True)
    @mock.patch(
        "uaclient.entitlements.fips.FIPSCommonEntitlement._perform_enable"
    )
    @mock.patch(M_PATH + "get_cloud_type")
    def test_show_message_on_cloud_id_error(
        self,
        m_get_cloud_type,
        m_perform_enable,
        caplog_text,
        entitlement_factory,
    ):
        m_get_cloud_type.return_value = (
            None,
            NoCloudTypeReason.CLOUD_ID_ERROR,
        )
        fips_entitlement = entitlement_factory(FIPSEntitlement)
        fips_entitlement._perform_enable(mock.MagicMock())
        logs = caplog_text()
        assert (
            "Could not determine cloud, defaulting to generic FIPS package."
            in logs
        )


class TestFIPSEntitlementRemovePackages:
    @pytest.mark.parametrize("installed_pkgs", (["sl"], ["ubuntu-fips", "sl"]))
    @mock.patch(M_PATH + "system.is_container", return_value=False)
    @mock.patch(
        "uaclient.system.get_release_info",
        return_value=mock.MagicMock(series="xenial"),
    )
    @mock.patch(M_PATH + "system.subp")
    @mock.patch(M_PATH + "apt.get_installed_packages_names")
    def test_remove_packages_only_removes_if_package_is_installed(
        self,
        m_get_installed_packages,
        m_subp,
        _m_get_release_info,
        _m_is_container,
        installed_pkgs,
        entitlement,
    ):
        m_subp.return_value = ("success", "")
        m_get_installed_packages.return_value = installed_pkgs
        entitlement.remove_packages()
        remove_cmd = mock.call(
            [
                "apt-get",
                "remove",
                "--assume-yes",
                '-o Dpkg::Options::="--force-confdef"',
                '-o Dpkg::Options::="--force-confold"',
                "ubuntu-fips",
            ],
            capture=True,
            retry_sleeps=apt.APT_RETRIES,
            override_env_vars={"DEBIAN_FRONTEND": "noninteractive"},
        )
        if "ubuntu-fips" in installed_pkgs:
            assert [remove_cmd] == m_subp.call_args_list
        else:
            assert [] == m_subp.call_args_list

    @mock.patch(
        "uaclient.system.get_release_info",
        return_value=mock.MagicMock(series="xenial"),
    )
    @mock.patch(M_PATH + "system.subp")
    @mock.patch(M_PATH + "apt.get_installed_packages_names")
    def test_remove_packages_output_message_when_fail(
        self,
        m_get_installed_packages,
        m_subp,
        _m_get_release_info,
        entitlement,
    ):
        m_get_installed_packages.return_value = ["ubuntu-fips"]
        m_subp.side_effect = exceptions.ProcessExecutionError(cmd="test")
        expected_msg = (
            "Unexpected APT error.\nCould not disable {}.\nSee "
            "/var/log/ubuntu-advantage.log"
        ).format(entitlement.title)

        with pytest.raises(exceptions.UbuntuProError) as exc_info:
            entitlement.remove_packages()

        assert exc_info.value.msg.strip() == expected_msg


class TestFIPSCheckForRebootMsg:
    @pytest.mark.parametrize(
        "operation,expected",
        (
            (
                "disable operation",
                messages.FIPS_DISABLE_REBOOT_REQUIRED,
            ),
        ),
    )
    @mock.patch("uaclient.system.should_reboot", return_value=True)
    def test_check_for_reboot_message(
        self,
        _m_should_reboot,
        operation,
        expected,
        entitlement,
    ):
        """When can_disable, disable removes apt config and packages."""
        notice_ent_cls = NoticesManager()
        entitlement._check_for_reboot_msg(operation=operation)
        assert [expected] == notice_ent_cls.list()


@mock.patch("uaclient.system.should_reboot")
class TestFIPSEntitlementApplicationStatus:
    @pytest.mark.parametrize(
        "super_application_status",
        [s for s in ApplicationStatus if s is not ApplicationStatus.ENABLED],
    )
    @mock.patch("uaclient.system.is_container", return_value=False)
    @mock.patch("os.path.exists", return_value=False)
    def test_non_enabled_passed_through(
        self,
        _m_path_exists,
        _m_is_container,
        _m_should_reboot,
        entitlement,
        super_application_status,
    ):
        msg = "sure is some status here"
        with mock.patch(
            M_PATH + "repo.RepoEntitlement.application_status",
            return_value=(super_application_status, msg),
        ):
            application_status = entitlement.application_status()

        assert (super_application_status, msg) == application_status

    @pytest.mark.parametrize(
        "super_application_status",
        [s for s in ApplicationStatus if s is not ApplicationStatus.ENABLED],
    )
    @mock.patch("uaclient.system.is_container", return_value=False)
    @mock.patch("os.path.exists", return_value=False)
    def test_non_root_does_not_fail(
        self,
        _m_path_exists,
        _m_is_container,
        _m_should_reboot,
        super_application_status,
        FakeConfig,
    ):
        cfg = FakeConfig()
        entitlement = FIPSUpdatesEntitlement(cfg)
        msg = "sure is some status here"
        with mock.patch(
            M_PATH + "repo.RepoEntitlement.application_status",
            return_value=(super_application_status, msg),
        ):
            application_status = entitlement.application_status()

        assert (super_application_status, msg) == application_status

    @pytest.mark.parametrize("should_reboot", ((True), (False)))
    @pytest.mark.parametrize("path_exists", ((True), (False)))
    @pytest.mark.parametrize("proc_content", (("0"), ("1")))
    @mock.patch(
        M_PATH + "apt.get_installed_packages_names",
        return_value=["ubuntu-fips"],
    )
    @mock.patch("uaclient.system.is_container", return_value=False)
    def test_proc_file_is_used_to_determine_application_status_message(
        self,
        _m_is_container,
        _m_installed_packages,
        m_should_reboot,
        proc_content,
        path_exists,
        should_reboot,
        entitlement,
    ):
        notice_ent_cls = NoticesManager()
        m_should_reboot.return_value = should_reboot
        orig_load_file = system.load_file

        def fake_load_file(path):
            if path == "/proc/sys/crypto/fips_enabled":
                return proc_content
            return orig_load_file(path)

        orig_exists = os.path.exists

        def fake_exists(path):
            if path == "/proc/sys/crypto/fips_enabled":
                return path_exists
            return orig_exists(path)

        msg = messages.NamedMessage("test-code", "sure is some status here")
        notice_ent_cls.add(
            Notice.FIPS_SYSTEM_REBOOT_REQUIRED,
            messages.FIPS_SYSTEM_REBOOT_REQUIRED,
        )

        if path_exists:
            notice_ent_cls.add(
                Notice.FIPS_REBOOT_REQUIRED,
                messages.FIPS_REBOOT_REQUIRED_MSG,
            )

        if proc_content == "0":
            notice_ent_cls.add(
                Notice.FIPS_DISABLE_REBOOT_REQUIRED,
                messages.FIPS_DISABLE_REBOOT_REQUIRED,
            )

        with mock.patch(
            M_PATH + "repo.RepoEntitlement.application_status",
            return_value=(ApplicationStatus.ENABLED, msg),
        ):
            with mock.patch("uaclient.system.load_file") as m_load_file:
                m_load_file.side_effect = fake_load_file
                with mock.patch("os.path.exists") as m_path_exists:
                    m_path_exists.side_effect = fake_exists
                    (
                        actual_status,
                        actual_msg,
                    ) = entitlement.application_status()

        expected_status = ApplicationStatus.ENABLED
        expected_msg = msg
        if path_exists and should_reboot and proc_content == "1":
            expected_msg = msg
            assert [
                messages.FIPS_SYSTEM_REBOOT_REQUIRED,
                messages.FIPS_REBOOT_REQUIRED_MSG,
            ] == notice_ent_cls.list()
        elif path_exists and not should_reboot and proc_content == "1":
            expected_msg = msg
            # we do not delete the FIPS_REBOOT_REQUIRED notices
            # deleting will happen after rebooting
            assert notice_ent_cls.list() == [messages.FIPS_REBOOT_REQUIRED_MSG]
        elif path_exists and should_reboot and proc_content == "0":
            expected_msg = messages.FIPS_PROC_FILE_ERROR.format(
                file_name=entitlement.FIPS_PROC_FILE
            )
            expected_status = ApplicationStatus.DISABLED
            assert [
                messages.FIPS_DISABLE_REBOOT_REQUIRED,
                messages.NOTICE_FIPS_MANUAL_DISABLE_URL,
                messages.FIPS_SYSTEM_REBOOT_REQUIRED,
                messages.FIPS_REBOOT_REQUIRED_MSG,
            ] == notice_ent_cls.list()
        elif path_exists and not should_reboot and proc_content == "0":
            expected_msg = messages.FIPS_PROC_FILE_ERROR.format(
                file_name=entitlement.FIPS_PROC_FILE
            )
            expected_status = ApplicationStatus.DISABLED
            assert [
                (
                    Notice.FIPS_MANUAL_DISABLE_URL.value.label,
                    messages.NOTICE_FIPS_MANUAL_DISABLE_URL,
                )
                == notice_ent_cls.list()
            ]
        else:
            expected_msg = messages.FIPS_REBOOT_REQUIRED

        assert actual_status == expected_status
        assert expected_msg.msg == actual_msg.msg
        assert expected_msg.name == actual_msg.name

    @mock.patch("uaclient.system.is_container", return_value=False)
    @mock.patch(M_PATH + "os.path.exists", return_value=False)
    @mock.patch(
        M_PATH + "repo.RepoEntitlement.application_status",
        return_value=(
            ApplicationStatus.ENABLED,
            messages.NamedMessage("test-code", "sure is some status here"),
        ),
    )
    @mock.patch(
        M_PATH + "apt.get_installed_packages_names",
        return_value=["other", "packages", "not-fips"],
    )
    def test_warning_when_metapackage_not_installed(
        self,
        _m_installed_packages,
        _m_application_status,
        _m_path_exists,
        _m_is_container,
        _m_should_reboot,
        entitlement,
    ):
        status, msg = entitlement.application_status()
        assert ApplicationStatus.WARNING == status
        assert (
            messages.FIPS_PACKAGES_NOT_INSTALLED.format(
                packages="ubuntu-fips", service=entitlement.name
            )
            == msg
        )

    @mock.patch("uaclient.system.is_container", return_value=True)
    def test_application_status_inside_container(
        self,
        _m_is_container,
        m_should_reboot,
        entitlement,
    ):
        m_should_reboot.return_value = False
        notice_ent_cls = NoticesManager()
        notice_ent_cls.add(
            Notice.FIPS_SYSTEM_REBOOT_REQUIRED,
            messages.FIPS_SYSTEM_REBOOT_REQUIRED,
        )
        expected_status = ApplicationStatus.ENABLED
        expected_msg = "test"

        with mock.patch(
            M_PATH + "repo.RepoEntitlement.application_status",
            return_value=(
                expected_status,
                messages.NamedMessage("test", expected_msg),
            ),
        ):
            actual_status, actual_message = entitlement.application_status()

        assert expected_status == actual_status
        assert expected_msg == actual_message.msg
        assert [] == notice_ent_cls.list()

    @mock.patch(
        M_PATH + "apt.get_installed_packages_names",
        return_value=["ubuntu-fips"],
    )
    def test_fips_does_not_show_enabled_when_fips_updates_is(
        self, _m_installed_packages, _m_should_reboot, entitlement
    ):
        with mock.patch("uaclient.apt.get_apt_cache_policy") as m_apt_policy:
            m_apt_policy.return_value = (
                "1001 http://FIPS-UPDATES/ubuntu"
                " xenial-updates/main amd64 Packages\n"
                ""
            )

            application_status, _ = entitlement.application_status()

        expected_status = ApplicationStatus.DISABLED
        if isinstance(entitlement, FIPSUpdatesEntitlement):
            expected_status = ApplicationStatus.ENABLED

        assert expected_status == application_status


class TestFipsEntitlementInstallPackages:
    @pytest.mark.parametrize(
        [
            "mandatory_packages",
            "auto_upgrade_all_on_enable",
            "check_for_reboot_result",
            "expected_super_install_packages_calls",
            "expected_install_all_available_upgrades_calls",
            "expected_hardcoded_install_conditional_packages_calls",
            "expected_notices_add_calls",
        ],
        [
            # no mandatory packages, hardcoded upgrades, no reboot
            (
                [],
                False,
                False,
                [],
                [],
                [mock.call(mock.ANY)],
                [],
            ),
            # mandatory packages, hardcoded upgrades, no reboot
            (
                ["ubuntu-fips"],
                False,
                False,
                [mock.call(mock.ANY, package_list=["ubuntu-fips"])],
                [],
                [mock.call(mock.ANY)],
                [],
            ),
            # mandatory packages, auto upgrades, no reboot
            (
                ["ubuntu-fips"],
                True,
                False,
                [mock.call(mock.ANY, package_list=["ubuntu-fips"])],
                [mock.call(mock.ANY)],
                [],
                [],
            ),
            # mandatory packages, auto upgrades, reboot
            (
                ["ubuntu-fips"],
                True,
                True,
                [mock.call(mock.ANY, package_list=["ubuntu-fips"])],
                [mock.call(mock.ANY)],
                [],
                [mock.call(mock.ANY)],
            ),
        ],
    )
    @mock.patch(M_PATH + "notices.add")
    @mock.patch(M_PATH + "FIPSCommonEntitlement._check_for_reboot")
    @mock.patch(
        M_PATH + "FIPSCommonEntitlement.hardcoded_install_conditional_packages"
    )
    @mock.patch(
        M_PATH + "FIPSCommonEntitlement.install_all_available_fips_upgrades"
    )
    @mock.patch(M_PATH + "FIPSCommonEntitlement.auto_upgrade_all_on_enable")
    @mock.patch(M_PATH + "repo.RepoEntitlement.install_packages")
    @mock.patch(
        M_PATH + "FIPSCommonEntitlement.packages",
        new_callable=mock.PropertyMock,
    )
    def test_install_packages(
        self,
        m_packages,
        m_install_packages,
        m_auto_upgrade_all_on_enable,
        m_install_all_available_fips_upgrades,
        m_hardcoded_install_conditional_packages,
        m_check_for_reboot,
        m_notices_add,
        mandatory_packages,
        auto_upgrade_all_on_enable,
        check_for_reboot_result,
        expected_super_install_packages_calls,
        expected_install_all_available_upgrades_calls,
        expected_hardcoded_install_conditional_packages_calls,
        expected_notices_add_calls,
        FakeConfig,
    ):
        m_packages.return_value = mandatory_packages
        m_auto_upgrade_all_on_enable.return_value = auto_upgrade_all_on_enable
        m_check_for_reboot.return_value = check_for_reboot_result

        ConcreteFIPSCommonEntitlement(FakeConfig()).install_packages(
            mock.MagicMock()
        )

        assert (
            expected_super_install_packages_calls
            == m_install_packages.call_args_list
        )
        assert (
            expected_install_all_available_upgrades_calls
            == m_install_all_available_fips_upgrades.call_args_list
        )
        assert (
            expected_hardcoded_install_conditional_packages_calls
            == m_hardcoded_install_conditional_packages.call_args_list
        )
        assert expected_notices_add_calls == m_notices_add.call_args_list

    @pytest.mark.parametrize(
        [
            "series",
            "to_upgrade",
            "get_pkg_candidate_version_side_effect",
            "expected_unhold_calls",
            "expected_run_apt_install_command_calls",
        ],
        [
            # no packages to upgrade
            (
                "other",
                [],
                None,
                [],
                [],
            ),
            # some packages to upgrade, no -hmacs
            (
                "other",
                [
                    helpers.mock_with_name_attr(name="one"),
                    helpers.mock_with_name_attr(name="two"),
                    helpers.mock_with_name_attr(name="three"),
                ],
                [None, None, None],
                [
                    mock.call(
                        ["one", "three", "two"],
                    )
                ],
                [
                    mock.call(
                        packages=["one", "three", "two"],
                        override_env_vars=mock.ANY,
                        apt_options=mock.ANY,
                    )
                ],
            ),
            # Jammy special case
            # Hardcoded, to be removed soon TM
            (
                "jammy",
                [
                    helpers.mock_with_name_attr(name="one"),
                    helpers.mock_with_name_attr(name="two"),
                    helpers.mock_with_name_attr(name="three"),
                ],
                [None, None, None],
                [
                    mock.call(
                        ["one", "openssl-fips-module-3", "three", "two"],
                    )
                ],
                [
                    mock.call(
                        packages=[
                            "one",
                            "openssl-fips-module-3",
                            "three",
                            "two",
                        ],
                        override_env_vars=mock.ANY,
                        apt_options=mock.ANY,
                    )
                ],
            ),
        ],
    )
    @mock.patch(M_PATH + "apt.run_apt_install_command")
    @mock.patch(M_PATH + "FIPSCommonEntitlement.unhold_packages")
    @mock.patch(M_PATH + "apt.get_pkg_candidate_version")
    @mock.patch(
        M_PATH
        + "apt.get_installed_packages_with_uninstalled_candidate_in_origin"
    )
    @mock.patch(M_PATH + "system.get_release_info")
    def test_install_all_available_fips_upgrades(
        self,
        m_release_info,
        m_get_upgrades,
        m_get_pkg_candidate_version,
        m_unhold_packages,
        m_run_apt_install_command,
        series,
        to_upgrade,
        get_pkg_candidate_version_side_effect,
        expected_unhold_calls,
        expected_run_apt_install_command_calls,
        FakeConfig,
    ):
        m_release_info.return_value.series = series
        m_get_upgrades.return_value = to_upgrade
        m_get_pkg_candidate_version.side_effect = (
            get_pkg_candidate_version_side_effect
        )
        ConcreteFIPSCommonEntitlement(
            FakeConfig()
        ).install_all_available_fips_upgrades(mock.MagicMock())
        assert expected_unhold_calls == m_unhold_packages.call_args_list
        assert (
            expected_run_apt_install_command_calls
            == m_run_apt_install_command.call_args_list
        )

    @pytest.mark.parametrize(
        [
            "feature_flag",
            "series",
            "expected_result",
        ],
        [
            (
                False,
                "xenial",
                False,
            ),
            (
                False,
                "bionic",
                False,
            ),
            (
                False,
                "focal",
                False,
            ),
            (
                False,
                "jammy",
                True,
            ),
            (
                False,
                "noble",
                True,
            ),
            (
                True,
                "xenial",
                True,
            ),
        ],
    )
    @mock.patch(M_PATH + "system.get_release_info")
    @mock.patch(M_PATH + "util.is_config_value_true")
    def test_auto_upgrade_all_on_enable(
        self,
        m_is_config_value_true,
        m_get_release_info,
        feature_flag,
        series,
        expected_result,
        FakeConfig,
    ):
        m_is_config_value_true.return_value = feature_flag
        m_get_release_info.return_value = mock.MagicMock(series=series)
        assert (
            expected_result
            == ConcreteFIPSCommonEntitlement(
                FakeConfig()
            ).auto_upgrade_all_on_enable()
        )

    @pytest.mark.parametrize(
        [
            "installed_packages_names",
            "conditional_packages",
            "expected_run_apt_install_command_calls",
        ],
        [
            (
                ["one", "two", "three"],
                [],
                [],
            ),
            (
                ["one", "two", "three"],
                ["one"],
                [
                    mock.call(
                        packages=["one"],
                        override_env_vars=mock.ANY,
                        apt_options=mock.ANY,
                    ),
                ],
            ),
            (
                ["one", "two", "three"],
                ["one", "one-hmac", "three"],
                [
                    mock.call(
                        packages=["one"],
                        override_env_vars=mock.ANY,
                        apt_options=mock.ANY,
                    ),
                    mock.call(
                        packages=["one-hmac"],
                        override_env_vars=mock.ANY,
                        apt_options=mock.ANY,
                    ),
                    mock.call(
                        packages=["three"],
                        override_env_vars=mock.ANY,
                        apt_options=mock.ANY,
                    ),
                ],
            ),
        ],
    )
    @mock.patch(M_PATH + "apt.run_apt_install_command")
    @mock.patch(M_PATH + "apt.get_installed_packages_names")
    def test_hardcoded_install_conditional_packages(
        self,
        m_get_installed_packages_names,
        m_run_apt_install_command,
        installed_packages_names,
        conditional_packages,
        expected_run_apt_install_command_calls,
        FakeConfig,
    ):
        m_get_installed_packages_names.return_value = installed_packages_names
        ent = ConcreteFIPSCommonEntitlement(FakeConfig())
        ent.conditional_packages = conditional_packages
        ent.hardcoded_install_conditional_packages(mock.MagicMock())
        assert (
            expected_run_apt_install_command_calls
            == m_run_apt_install_command.call_args_list
        )


class TestFipsSetupAPTConfig:
    @pytest.mark.parametrize(
        "held_packages,unhold_packages",
        (
            ("", []),
            ("asdf\n", []),
            (
                "openssh-server\nlibssl1.1-hmac\nasdf\n",
                ["openssh-server", "libssl1.1-hmac"],
            ),
            (
                "libgcrypt20\nlibgcrypt20-hmac\nwow\n",
                ["libgcrypt20", "libgcrypt20-hmac"],
            ),
        ),
    )
    @mock.patch(M_REPOPATH + "RepoEntitlement.setup_apt_config")
    @mock.patch(M_PATH + "apt.run_apt_command")
    def test_setup_apt_cofig_unmarks_held_fips_packages(
        self,
        run_apt_command,
        setup_apt_config,
        held_packages,
        unhold_packages,
        entitlement,
    ):
        """Unmark only fips-specific package holds if present."""
        run_apt_command.return_value = held_packages
        entitlement.setup_apt_config(mock.MagicMock())
        expected_calls = [
            mock.call(
                ["apt-mark", "showholds"],
                messages.EXECUTING_COMMAND_FAILED.format(
                    command="apt-mark showholds"
                ),
            )
        ]
        if unhold_packages:
            cmd = ["apt-mark", "unhold"] + unhold_packages
            expected_calls.append(
                mock.call(
                    cmd,
                    messages.EXECUTING_COMMAND_FAILED.format(
                        command=" ".join(cmd)
                    ),
                )
            )
        assert expected_calls == run_apt_command.call_args_list
        assert [mock.call(mock.ANY)] == setup_apt_config.call_args_list


class TestFipsEntitlementPackages:
    @mock.patch(M_PATH + "apt.get_installed_packages_names", return_value=[])
    @mock.patch("uaclient.system.get_release_info")
    def test_packages_is_list(self, m_get_release_info, _mock, entitlement):
        """RepoEntitlement.enable will fail if it isn't"""

        # Do not trigger metapackage override by
        # _replace_metapackage_on_cloud_instance
        m_get_release_info.return_value = mock.MagicMock(series="test")

        assert isinstance(entitlement.packages, list)

    @mock.patch("uaclient.system.is_container", return_value=True)
    def test_packages_is_empty_if_running_on_container(
        self, _m_is_container, entitlement
    ):
        assert [] == entitlement.packages

    @mock.patch(M_PATH + "apt.get_installed_packages_names", return_value=[])
    @mock.patch("uaclient.system.get_release_info")
    def test_fips_required_packages_included(
        self, m_get_release_info, _mock, entitlement
    ):
        """The fips_required_packages should always be in .packages"""

        # Do not trigger metapackage override by
        # _replace_metapackage_on_cloud_instance
        m_get_release_info.return_value = mock.MagicMock(series="test")

        assert set(FIPS_ADDITIONAL_PACKAGES).issubset(
            set(entitlement.packages)
        )

    @mock.patch("uaclient.system.get_release_info")
    def test_currently_installed_packages_are_included_in_packages(
        self, m_get_release_info, entitlement
    ):
        # Do not trigger metapackage override by
        # _replace_metapackage_on_cloud_instance
        # and xenial should not trigger that
        m_get_release_info.return_value = mock.MagicMock(series="xenial")

        assert sorted(FIPS_ADDITIONAL_PACKAGES) == sorted(entitlement.packages)

    @mock.patch(M_PATH + "apt.get_installed_packages_names")
    @mock.patch("uaclient.system.get_release_info")
    def test_multiple_packages_calls_dont_mutate_state(
        self, m_get_release_info, m_get_installed_packages, entitlement
    ):
        # Make it appear like all packages are installed
        m_get_installed_packages.return_value.__contains__.return_value = True

        # Do not trigger metapackage override by
        # _replace_metapackage_on_cloud_instance
        m_get_release_info.return_value = mock.MagicMock(series="test")

        before = copy.deepcopy(entitlement.conditional_packages)

        assert entitlement.packages

        after = copy.deepcopy(entitlement.conditional_packages)

        assert before == after


class TestFIPSUpdatesEntitlementEnable:
    @pytest.mark.parametrize("enable_ret", ((True), (False)))
    @mock.patch("uaclient.entitlements.fips.notices.NoticesManager.remove")
    @mock.patch(
        "uaclient.entitlements.fips.FIPSCommonEntitlement._perform_enable"
    )
    def test_fips_updates_enable_write_service_once_enable_file(
        self,
        m_perform_enable,
        m_remove_notice,
        enable_ret,
        entitlement_factory,
    ):
        m_perform_enable.return_value = enable_ret
        fake_file = mock.MagicMock()
        fake_file.read.return_value = None

        with mock.patch.object(
            fips, "services_once_enabled_file", fake_file
        ) as m_services_once_enabled:
            cfg = mock.MagicMock()
            fips_updates_ent = entitlement_factory(
                FIPSUpdatesEntitlement, cfg=cfg
            )
            assert (
                fips_updates_ent._perform_enable(mock.MagicMock())
                == enable_ret
            )

        if enable_ret:
            assert 1 == m_services_once_enabled.write.call_count
        else:
            assert not m_services_once_enabled.write.call_count


class TestFIPSEntitlementCanEnable:
    @pytest.mark.parametrize(
        "fips_cls",
        (
            (FIPSUpdatesEntitlement),
            (FIPSPreviewEntitlement),
        ),
    )
    @mock.patch("uaclient.util.is_config_value_true", return_value=False)
    def test_can_enable_false_if_fips_enabled(
        self, m_is_config_value_true, fips_cls, capsys, entitlement_factory
    ):
        """When entitlement is disabled, can_enable returns True."""
        entitlement = entitlement_factory(fips_cls)
        with mock.patch.object(
            entitlement,
            "applicability_status",
            return_value=(ApplicabilityStatus.APPLICABLE, ""),
        ):
            with mock.patch.object(
                entitlement,
                "application_status",
                return_value=(ApplicationStatus.DISABLED, ""),
            ):
                with mock.patch(
                    M_PATH + "FIPSEntitlement.application_status"
                ) as m_fips_status:
                    m_fips_status.return_value = (
                        ApplicationStatus.ENABLED,
                        None,
                    )
                    actual_ret, reason = entitlement.can_enable()
                    assert actual_ret is False
                    assert (
                        reason.reason
                        == CanEnableFailureReason.INCOMPATIBLE_SERVICE
                    )


class TestFipsPreview:
    def test_allow_fips_on_cloud_is_always_true(self, entitlement_factory):
        entitlement = entitlement_factory(FIPSPreviewEntitlement)
        assert entitlement._allow_fips_on_cloud_instance(
            series="test", cloud_id="test"
        )



import json
import logging
import re
import textwrap

import jq  # type: ignore
import jsonschema  # type: ignore
import yaml
from behave import then, when
from hamcrest import assert_that, contains_string, equal_to, not_

from features.steps.shell import when_i_run_command
from features.util import SafeLoaderWithoutDatetime, process_template_vars


@then("I will see the following on {stream}:")
def then_i_will_see_on_stream(context, stream):
    content = getattr(context.process, stream).strip()
    text = process_template_vars(context, context.text)
    logging.debug("repr(expected): %r", text)
    logging.debug("repr(actual): %r", content)
    if not text == content:
        raise AssertionError(
            "Expected to find exactly:\n{}\nBut got:\n{}".format(
                textwrap.indent(text, "  "), textwrap.indent(content, "  ")
            )
        )


@then("if `{value1}` in `{value2}` I will see the following on {stream}:")
def then_conditional_i_will_see_on_stream(context, value1, value2, stream):
    """Only apply regex assertion if value1 in value2."""
    if value1 in value2.split(" or "):
        then_i_will_see_on_stream(context, stream)


@then("if `{value1}` not in `{value2}` I will see the following on {stream}:")
def then_not_in_conditional_stream_matches_regexp(
    context, value1, value2, stream
):
    """Only apply regex assertion if value1 in value2."""
    if value1 not in value2.split(" or "):
        then_i_will_see_on_stream(context, stream)


@then("if `{value1}` in `{value2}` and {stream} matches regexp:")
def then_conditional_stream_matches_regexp(context, value1, value2, stream):
    """Only apply regex assertion if value1 in value2."""
    if value1 in value2.split(" or "):
        then_stream_matches_regexp(context, stream)


@then("if `{value1}` in `{value2}` and {stream} contains substring:")
def then_conditional_stdout_contains_substring(
    context, value1, value2, stream
):
    if value1 in value2.split(" or "):
        then_stream_contains_substring(context, stream)


@then("if `{value1}` not in `{value2}` and {stream} contains substring:")
def then_not_in_conditional_stdout_contains_substring(
    context, value1, value2, stream
):
    if value1 not in value2.split(" or "):
        then_stream_contains_substring(context, stream)


@then("if `{value1}` in `{value2}` and stdout does not match regexp:")
def then_conditional_stdout_does_not_match_regexp(context, value1, value2):
    """Only apply regex assertion if value1 in value2."""
    if value1 in value2.split(" or "):
        then_stream_does_not_match_regexp(context, "stdout")


@then("if `{value1}` not in `{value2}` and stdout matches regexp:")
def then_not_in_conditional_stdout_does_not_match_regexp(
    context, value1, value2
):
    """Only apply regex assertion if value1 not in value2."""
    if value1 not in value2.split(" or "):
        then_stream_matches_regexp(context, "stdout")


@then("{stream} does not match regexp:")
def then_stream_does_not_match_regexp(context, stream):
    text = process_template_vars(context, context.text)
    content = getattr(context.process, stream).strip()
    if re.compile(text).search(content) is not None:
        raise AssertionError(
            "Expected to NOT match regexp:\n{}\nBut got:\n{}".format(
                textwrap.indent(text, "  "), textwrap.indent(content, "  ")
            )
        )


def compare_regexp(expected_regex, actual_output):
    if re.compile(expected_regex).search(actual_output) is None:
        raise AssertionError(
            "Expected to match regexp:\n{}\nBut got:\n{}".format(
                textwrap.indent(expected_regex, "  "),
                textwrap.indent(actual_output, "  "),
            )
        )


def process_api_data(context, api_key=None, escape=True):
    json_data = json.loads(context.process.stdout.strip())
    if escape:
        context.text = context.text.replace("[", "\\[")
        context.text = context.text.replace("]", "\\]")
        context.text = context.text.replace("(", "\\(")
        context.text = context.text.replace(")", "\\)")
        context.text = context.text.replace("\\n", "\\\\n")

    if api_key is None:
        return json.dumps(json_data, indent=2)
    else:
        return json.dumps(json_data[api_key], indent=2)


@then("API full output matches regexp:")
def then_api_output_matches_regexp(context):
    content = process_api_data(context)
    text = process_template_vars(context, context.text)
    compare_regexp(text, content)


@then("API data field output matches regexp:")
def then_api_data_output_matches_regexp(context):
    content = process_api_data(context, api_key="data")
    text = process_template_vars(context, context.text)
    compare_regexp(text, content)


@then("API data field output is:")
def then_api_data_output_is(context):
    content = process_api_data(context, api_key="data", escape=False)
    text = process_template_vars(context, context.text)
    if not text == content:
        raise AssertionError(
            "Expected to find exactly:\n{}\nBut got:\n{}".format(
                textwrap.indent(text, "  "), textwrap.indent(content, "  ")
            )
        )


@then("API {field} field output matches regexp:")
def then_api_field_output_matches_regexp(context, field):
    content = process_api_data(context, api_key=field)
    text = process_template_vars(context, context.text)
    compare_regexp(text, content)


@then("API {field} field output is:")
def then_api_errors_output_is(context, field):
    content = process_api_data(context, api_key=field, escape=False)
    text = process_template_vars(context, context.text)
    if not text == content:
        raise AssertionError(
            "Expected to find exactly:\n{}\nBut got:\n{}".format(
                textwrap.indent(text, "  "), textwrap.indent(content, "  ")
            )
        )


@then("{stream} matches regexp:")
def then_stream_matches_regexp(context, stream):
    content = getattr(context.process, stream).strip()
    text = process_template_vars(context, context.text)
    compare_regexp(text, content)


@then("{stream} contains substring:")
def then_stream_contains_substring(context, stream):
    content = getattr(context.process, stream).strip()
    text = process_template_vars(context, context.text)
    if text not in content:
        raise AssertionError(
            (
                "Expected to find substring:\n{}\n"
                + "But couldn't find it in:\n{}"
            ).format(
                textwrap.indent(text, "  "), textwrap.indent(content, "  ")
            )
        )


@then("{stream} does not contain substring:")
def then_stream_not_contains_substring(context, stream):
    content = getattr(context.process, stream).strip()
    text = process_template_vars(context, context.text)
    assert_that(content, not_(contains_string(text)))
    if text in content:
        raise AssertionError(
            (
                "Expected to NOT find substring:\n{}\n"
                + "But did find it in:\n{}"
            ).format(
                textwrap.indent(text, "  "), textwrap.indent(content, "  ")
            )
        )


@then("I will see the uaclient version on stdout")
def then_i_will_see_the_uaclient_version_on_stdout(context):
    python_import = "from uaclient.version import get_version"

    cmd = "python3 -c '{}; print(get_version())'".format(python_import)

    actual_version = context.process.stdout.strip()
    when_i_run_command(context, cmd, "as non-root")
    expected_version = context.process.stdout.strip()

    assert_that(expected_version, equal_to(actual_version))


@then("stdout is a {output_format} matching the `{schema}` schema")
def stdout_matches_the_json_schema(context, output_format, schema):
    if output_format == "json":
        instance = json.loads(context.process.stdout.strip())
    elif output_format == "yaml":
        instance = yaml.load(
            context.process.stdout.strip(), SafeLoaderWithoutDatetime
        )
    with open("features/schemas/{}.json".format(schema), "r") as schema_file:
        jsonschema.validate(instance=instance, schema=json.load(schema_file))


@then("the {output_format} API response data matches the `{schema}` schema")
def api_response_matches_schema(context, output_format, schema):
    if output_format == "json":
        instance = json.loads(context.process.stdout.strip())
    elif output_format == "yaml":
        instance = yaml.load(
            context.process.stdout.strip(), SafeLoaderWithoutDatetime
        )
    with open("features/schemas/{}.json".format(schema), "r") as schema_file:
        jsonschema.validate(
            instance=instance.get("data", {}).get("attributes"),
            schema=json.load(schema_file),
        )


@when("I verify root and non-root `{cmd}` calls have the same output")
def root_vs_nonroot_cmd_comparison(context, cmd):
    when_i_run_command(context, cmd, "with sudo")
    root_status_stdout = context.process.stdout.strip()
    root_status_stderr = context.process.stderr.strip()

    when_i_run_command(context, cmd, "as non-root")
    nonroot_status_stdout = context.process.stdout.strip()
    nonroot_status_stderr = context.process.stderr.strip()

    assert_that(root_status_stdout, nonroot_status_stdout)
    assert root_status_stderr == nonroot_status_stderr


@when("I verify that `{field}` field is redacted in the logs")
def i_verify_field_is_redacted_in_the_logs(context, field):
    when_i_run_command(
        context, "cat /var/log/ubuntu-advantage.log", "with sudo"
    )
    context.text = field + "<REDACTED>"
    then_stream_contains_substring(context, "stdout")


@when("I apply this jq filter `{jq_filter}` to the output")
def i_apply_jq_filter(context, jq_filter):
    context.process.stdout = json.dumps(
        json.loads(
            jq.compile(jq_filter)
            .input_text(context.process.stdout.strip())
            .text()
        ),
        indent=2,
    )


@when("I apply this jq filter `{jq_filter}` to the API data field output")
def i_apply_jq_filter_to_api_data(context, jq_filter):
    content = process_api_data(context, api_key="data", escape=False)
    context.process.stdout = jq.compile(jq_filter).input_text(content).text()


@when("I remove colors from output")
def i_remove_colors_from_output(context):
    COLOR_FORMATTING_PATTERN = r"\033\[.*?m"
    context.process.stdout = re.sub(
        COLOR_FORMATTING_PATTERN, "", context.process.stdout.strip()
    )


@when("I remove links from output")
def i_remove_links_from_output(context):
    LINK_START_PATTERN = r"\033]8;;.+?\033\\+"
    LINK_END = "\033]8;;\033\\"
    context.process.stdout = re.sub(
        LINK_START_PATTERN, "", context.process.stdout.strip()
    )
    context.process.stdout = context.process.stdout.replace(LINK_END, "")



from typing import Any, Callable, Dict, List, Optional, Tuple, Union

from uaclient.messages import NamedMessage

StaticAffordance = Tuple[NamedMessage, Callable[[], Any], bool]

MessagingOperations = List[Union[str, Tuple[Callable, Dict]]]
MessagingOperationsDict = Dict[str, Optional[MessagingOperations]]



import re
import subprocess
import sys

sys.path.insert(0, ".")

from uaclient.version import __VERSION__

python_version = __VERSION__
changelog_version = (
    subprocess.check_output(["dpkg-parsechangelog", "-S", "version"])
    .decode("utf-8")
    .strip()
)

# We care about consistent versioning when releasing,
# skip checking development versions
if "+devel" in changelog_version:
    sys.exit(0)

# remove tilde and suffix of changelog_version if present
# typical version strings:
# GH: 32.3ubuntu1~1.gbp761c11~noble1 -> 32.3
# backports: 32.3~22.04 -> 32.3
# ppa test builds: 32.3~22.04~ppa1 -> 32.3

m = re.match(r"(\d+(\.\d+)*)", changelog_version)
if m:
    base_changelog_version = m.group()
else:
    base_changelog_version = None

if python_version != base_changelog_version:
    print(
        'version.py says "{}" but changelog says "{}" (base version: "{}")'.format(
            python_version, changelog_version, base_changelog_version
        ),
        file=sys.stderr,
    )
    sys.exit(1)



from typing import Any, Dict, List, Optional

import pytest

from uaclient import config, event_logger


def machine_token(
    entitlement_type: str,
    *,
    affordances: Optional[Dict[str, Any]] = None,
    directives: Optional[Dict[str, Any]] = None,
    overrides: Optional[List[Dict[str, Any]]] = None,
    entitled: bool = True,
    obligations: Optional[Dict[str, Any]] = None,
    suites: Optional[List[str]] = None,
    additional_packages: Optional[List[str]] = None
) -> Dict[str, Any]:
    return {
        "resourceTokens": [
            {
                "type": entitlement_type,
                "token": "{}-token".format(entitlement_type),
            }
        ],
        "machineToken": "blah",
        "machineTokenInfo": {
            "contractInfo": {
                "resourceEntitlements": [
                    machine_access(
                        entitlement_type,
                        affordances=affordances,
                        directives=directives,
                        overrides=overrides,
                        entitled=entitled,
                        obligations=obligations,
                        suites=suites,
                        additional_packages=additional_packages,
                    )
                ]
            }
        },
    }


def machine_access(
    entitlement_type: str,
    *,
    affordances: Optional[Dict[str, Any]] = None,
    directives: Optional[Dict[str, Any]] = None,
    overrides: Optional[List[Dict[str, Any]]] = None,
    entitled: bool = True,
    obligations: Optional[Dict[str, Any]] = None,
    suites: Optional[List[str]] = None,
    additional_packages: Optional[List[str]] = None
) -> Dict[str, Any]:
    if affordances is None:
        affordances = {}
    if suites is None:
        suites = ["xenial"]
    if obligations is None:
        obligations = {"enableByDefault": True}
    if directives is None:
        directives = {
            "aptURL": "http://{}".format(entitlement_type.upper()),
            "aptKey": "APTKEY",
            "suites": suites,
        }

        if additional_packages:
            directives["additionalPackages"] = additional_packages
    if overrides is None:
        overrides = []

    return {
        "obligations": obligations,
        "type": entitlement_type,
        "entitled": entitled,
        "directives": directives,
        "affordances": affordances,
        "overrides": overrides,
    }


@pytest.fixture
def entitlement_factory(tmpdir, FakeConfig, fake_machine_token_file):
    """
    A pytest fixture that returns a function that instantiates an entitlement

    The function requires an entitlement class as its first argument, and takes
    keyword arguments for affordances, directives and suites which, if given,
    replace the default values in the resourceEntitlements of the
    machine-token.json file for the entitlement.
    """

    def factory_func(
        cls,
        *,
        affordances: Optional[Dict[str, Any]] = None,
        directives: Optional[Dict[str, Any]] = None,
        obligations: Optional[Dict[str, Any]] = None,
        overrides: Optional[List[Dict[str, Any]]] = None,
        entitled: bool = True,
        called_name: str = "",
        access_only: bool = False,
        purge: bool = False,
        suites: Optional[List[str]] = None,
        additional_packages: Optional[List[str]] = None,
        cfg: Optional[config.UAConfig] = None,
        cfg_extension: Optional[Dict[str, Any]] = None,
        cfg_features: Optional[Dict[str, Any]] = None,
        # Those extra args should be used for scenarios where a cls
        # instance requires something that is not shared between all
        # entitlement classes
        extra_args: Optional[Dict[str, Any]] = None
    ):
        if not cfg:
            cfg_arg = {"data_dir": tmpdir.strpath}
            if cfg_extension is not None:
                cfg_arg.update(cfg_extension)
            if cfg_features is not None:
                cfg_arg["features"] = cfg_features
            cfg = FakeConfig(cfg_overrides=cfg_arg)
            fake_machine_token_file.attached = True
            fake_machine_token_file.token = machine_token(
                cls.name,
                affordances=affordances,
                directives=directives,
                overrides=overrides,
                obligations=obligations,
                entitled=entitled,
                suites=suites,
                additional_packages=additional_packages,
            )

        args = {
            "called_name": called_name,
            "access_only": access_only,
            "purge": purge,
        }

        if extra_args:
            args = {**args, **extra_args}

        return cls(cfg, **args)

    return factory_func


@pytest.fixture
def event():
    event = event_logger.get_event_logger()
    event.reset()

    return event



import mock
import pytest

from uaclient import messages
from uaclient.api.u.pro.security.fix._common import FixStatus
from uaclient.api.u.pro.security.fix._common.execute.v1 import (
    FailedUpgrade,
    FixExecuteError,
    FixExecuteResult,
    UpgradedPackage,
    _execute_fix,
)
from uaclient.api.u.pro.security.fix._common.plan.v1 import (
    ESM_INFRA_POCKET,
    STANDARD_UPDATES_POCKET,
    AptUpgradeData,
    AttachData,
    EnableData,
    FixPlanAptUpgradeStep,
    FixPlanAttachStep,
    FixPlanEnableStep,
    FixPlanError,
    FixPlanNoOpStatus,
    FixPlanNoOpStep,
    FixPlanResult,
    FixPlanWarningPackageCannotBeInstalled,
    FixPlanWarningSecurityIssueNotFixed,
    FixWarningType,
    NoOpData,
    PackageCannotBeInstalledData,
    SecurityIssueNotFixedData,
)


class TestExecute:
    @pytest.mark.parametrize(
        "security_issue,expected_result,pkg_version",
        (
            # Attach step
            (
                FixPlanResult(
                    title="CVE-12345-567",
                    description="description",
                    current_status="still-affected",
                    expected_status=FixStatus.SYSTEM_NON_VULNERABLE.value.msg,
                    affected_packages=["pkg1", "pkg2"],
                    plan=[
                        FixPlanAttachStep(
                            data=AttachData(
                                reason="test",
                                required_service=ESM_INFRA_POCKET,
                                source_packages=["pkg1", "pkg2"],
                            ),
                            order=1,
                        ),
                        FixPlanEnableStep(
                            data=EnableData(
                                service=ESM_INFRA_POCKET,
                                source_packages=["pkg1", "pkg2"],
                            ),
                            order=2,
                        ),
                        FixPlanAptUpgradeStep(
                            data=AptUpgradeData(
                                binary_packages=["pkg1", "pkg2"],
                                source_packages=["pkg1", "pkg2"],
                                pocket=ESM_INFRA_POCKET,
                            ),
                            order=3,
                        ),
                    ],
                    warnings=None,
                    error=None,
                    additional_data=None,
                ),
                FixExecuteResult(
                    title="CVE-12345-567",
                    description="description",
                    status=FixStatus.SYSTEM_STILL_VULNERABLE.value.msg,
                    upgraded_packages=[],
                    errors=[
                        FixExecuteError(
                            error_type="fix-requires-attach",
                            reason=messages.SECURITY_UPDATE_NOT_INSTALLED_SUBSCRIPTION,  # noqa
                            failed_upgrades=[
                                FailedUpgrade(
                                    name="pkg1", pocket=ESM_INFRA_POCKET
                                ),
                                FailedUpgrade(
                                    name="pkg2", pocket=ESM_INFRA_POCKET
                                ),
                            ],
                        )
                    ],
                ),
                None,
            ),
            # Enable step without attach
            (
                FixPlanResult(
                    title="CVE-12345-567",
                    description="description",
                    current_status="still-affected",
                    expected_status=FixStatus.SYSTEM_NON_VULNERABLE.value.msg,
                    affected_packages=["pkg1", "pkg2"],
                    plan=[
                        FixPlanEnableStep(
                            data=EnableData(
                                service=ESM_INFRA_POCKET,
                                source_packages=["pkg1", "pkg2"],
                            ),
                            order=1,
                        ),
                        FixPlanAptUpgradeStep(
                            data=AptUpgradeData(
                                binary_packages=["pkg1", "pkg2"],
                                source_packages=["pkg1", "pkg2"],
                                pocket=ESM_INFRA_POCKET,
                            ),
                            order=2,
                        ),
                    ],
                    warnings=None,
                    error=None,
                    additional_data=None,
                ),
                FixExecuteResult(
                    title="CVE-12345-567",
                    description="description",
                    status=FixStatus.SYSTEM_STILL_VULNERABLE.value.msg,
                    upgraded_packages=[],
                    errors=[
                        FixExecuteError(
                            error_type="fix-requires-enable",
                            reason=messages.SECURITY_SERVICE_DISABLED.format(
                                service=ESM_INFRA_POCKET
                            ),
                            failed_upgrades=[
                                FailedUpgrade(
                                    name="pkg1", pocket=ESM_INFRA_POCKET
                                ),
                                FailedUpgrade(
                                    name="pkg2", pocket=ESM_INFRA_POCKET
                                ),
                            ],
                        )
                    ],
                ),
                None,
            ),
            # Package cannot be installed warning
            (
                FixPlanResult(
                    title="CVE-12345-567",
                    description="description",
                    current_status="still-affected",
                    expected_status=FixStatus.SYSTEM_NON_VULNERABLE.value.msg,
                    affected_packages=["pkg1", "pkg2"],
                    plan=[
                        FixPlanAptUpgradeStep(
                            data=AptUpgradeData(
                                binary_packages=["bin2"],
                                source_packages=["pkg2"],
                                pocket=STANDARD_UPDATES_POCKET,
                            ),
                            order=2,
                        ),
                    ],
                    warnings=[
                        FixPlanWarningPackageCannotBeInstalled(
                            data=PackageCannotBeInstalledData(
                                binary_package="bin1",
                                binary_package_version="ver1",
                                source_package="pkg1",
                                related_source_packages=["pkg1", "pkg2"],
                                pocket=STANDARD_UPDATES_POCKET,
                            ),
                            order=1,
                        )
                    ],
                    error=None,
                    additional_data=None,
                ),
                FixExecuteResult(
                    title="CVE-12345-567",
                    description="description",
                    status=FixStatus.SYSTEM_STILL_VULNERABLE.value.msg,
                    upgraded_packages=[
                        UpgradedPackage(
                            name="bin2",
                            version="ver2",
                            pocket=STANDARD_UPDATES_POCKET,
                        )
                    ],
                    errors=[
                        FixExecuteError(
                            error_type=FixWarningType.PACKAGE_CANNOT_BE_INSTALLED.value,  # noqa
                            reason=messages.FIX_CANNOT_INSTALL_PACKAGE.format(
                                package="bin1",
                                version="ver1",
                            ),
                            failed_upgrades=[
                                FailedUpgrade(
                                    name="bin1", pocket=STANDARD_UPDATES_POCKET
                                ),
                            ],
                        )
                    ],
                ),
                "ver2",
            ),
            # Security issue not fixed warning
            (
                FixPlanResult(
                    title="CVE-12345-567",
                    description="description",
                    current_status="still-affected",
                    expected_status=FixStatus.SYSTEM_NON_VULNERABLE.value.msg,
                    affected_packages=["pkg1", "pkg2"],
                    plan=[
                        FixPlanAptUpgradeStep(
                            data=AptUpgradeData(
                                binary_packages=["bin2"],
                                source_packages=["pkg2"],
                                pocket=STANDARD_UPDATES_POCKET,
                            ),
                            order=2,
                        ),
                    ],
                    warnings=[
                        FixPlanWarningSecurityIssueNotFixed(
                            data=SecurityIssueNotFixedData(
                                source_packages=["pkg1"], status="pending"
                            ),
                            order=1,
                        )
                    ],
                    error=None,
                    additional_data=None,
                ),
                FixExecuteResult(
                    title="CVE-12345-567",
                    description="description",
                    status=FixStatus.SYSTEM_STILL_VULNERABLE.value.msg,
                    upgraded_packages=[
                        UpgradedPackage(
                            name="bin2",
                            version="ver2",
                            pocket=STANDARD_UPDATES_POCKET,
                        )
                    ],
                    errors=[
                        FixExecuteError(
                            error_type=FixWarningType.SECURITY_ISSUE_NOT_FIXED.value,  # noqa
                            reason=messages.SECURITY_CVE_STATUS_PENDING,
                            failed_upgrades=[
                                FailedUpgrade(name="pkg1"),
                            ],
                        )
                    ],
                ),
                "ver2",
            ),
            # CVE error
            (
                FixPlanResult(
                    title="CVE-1",
                    description=None,
                    current_status=None,
                    expected_status="error",
                    affected_packages=None,
                    plan=[
                        FixPlanAptUpgradeStep(
                            data=AptUpgradeData(
                                binary_packages=["bin2"],
                                source_packages=["pkg2"],
                                pocket=STANDARD_UPDATES_POCKET,
                            ),
                            order=2,
                        ),
                    ],
                    warnings=[],
                    error=FixPlanError(
                        code="security-issue-not-found-issue",
                        msg=messages.E_SECURITY_FIX_NOT_FOUND_ISSUE.format(
                            issue_id="CVE-"
                        ),
                    ),
                    additional_data=None,
                ),
                FixExecuteResult(
                    title="CVE-1",
                    description=None,
                    status="error",
                    upgraded_packages=[],
                    errors=[
                        FixExecuteError(
                            error_type="security-issue-not-found-issue",
                            reason=messages.E_SECURITY_FIX_NOT_FOUND_ISSUE.format(  # noqa
                                issue_id="CVE-"
                            ),
                        )
                    ],
                ),
                None,
            ),
            # Noop step
            (
                FixPlanResult(
                    title="CVE-12345-567",
                    description="description",
                    current_status="not-affected",
                    expected_status=FixStatus.SYSTEM_NOT_AFFECTED.value.msg,
                    affected_packages=["pkg1", "pkg2"],
                    plan=[
                        FixPlanNoOpStep(
                            data=NoOpData(
                                status=FixPlanNoOpStatus.NOT_AFFECTED.value
                            ),
                            order=1,
                        )
                    ],
                    warnings=[],
                    error=None,
                    additional_data=None,
                ),
                FixExecuteResult(
                    title="CVE-12345-567",
                    description="description",
                    status=FixStatus.SYSTEM_NOT_AFFECTED.value.msg,
                    upgraded_packages=[],
                    errors=None,
                ),
                None,
            ),
        ),
    )
    @mock.patch("uaclient.apt.run_apt_update_command")
    @mock.patch("uaclient.apt.run_apt_command")
    @mock.patch("uaclient.apt.get_pkg_version")
    def test_execute(
        self,
        m_get_pkg_version,
        _m_run_apt_cmd,
        _m_run_apt_update,
        security_issue,
        expected_result,
        pkg_version,
    ):
        m_get_pkg_version.return_value = pkg_version
        assert expected_result == _execute_fix(security_issue)



import os
import re
from typing import Any, Dict  # noqa: F401

import yaml
from behave import when

from features.steps.files import when_i_create_file_with_content
from features.steps.shell import when_i_run_command


@when("I download the service credentials on the `{machine_name}` machine")
def download_service_credentials(context, machine_name):
    token = context.pro_config.contract_token
    when_i_run_command(
        context,
        "get-resource-tokens {}".format(token),
        "as non-root",
        machine_name=machine_name,
    )

    context.service_credentials = context.process.stdout


@when(
    "I extract the `{service}` credentials from the `{machine_name}` machine"
)
def extract_service_credentials(context, service, machine_name):
    if getattr(context, "service_mirror_cfg", None) is None:
        context.service_mirror_cfg = {}
    credentials_pattern = r"{}:\s+(.+)".format(service)
    credentials_match = re.search(
        credentials_pattern, context.service_credentials
    )
    if credentials_match:
        extracted_credentials = credentials_match.group(1)
        context.service_mirror_cfg[service.replace("-", "_")] = {
            "credentials": extracted_credentials
        }


@when(
    "I set the apt-mirror file for `{release}` with the `{service_list}` credentials on the `{machine_name}` machine"  # noqa
)
def set_apt_mirror_file_with_credentials(
    context, release, service_list, machine_name
):
    service_list = service_list.split(",")

    apt_mirror_file = """
    set nthreads     20
    set _tilde 0
    """

    for service in service_list:
        token = context.service_mirror_cfg[service.replace("-", "_")][
            "credentials"
        ]
        service_type = service.split("-")[1]

        apt_mirror_cfg = """
        deb https://bearer:{}@esm.ubuntu.com/{}/ubuntu/ jammy-{}-updates main
        deb https://bearer:{}@esm.ubuntu.com/{}/ubuntu/ jammy-{}-security main
        """.format(
            token,
            service_type,
            service_type,
            token,
            service_type,
            service_type,
        )

        apt_mirror_file += "\n" + apt_mirror_cfg + "\n"
        context.service_mirror_cfg[service.replace("-", "_")]["path"] = (
            "/var/spool/apt-mirror/mirror/esm.ubuntu.com/{}/".format(
                service_type
            )
        )

    apt_mirror_file += "clean http://archive.ubuntu.com/ubuntu"

    context.text = apt_mirror_file.strip()
    when_i_create_file_with_content(
        context,
        "/etc/apt/mirror.list",
        machine_name=machine_name,
    )


@when(
    "I serve the `{service}` mirror using port `{port}` on the `{machine_name}` machine"  # noqa
)
def serve_apt_mirror(context, service, port, machine_name):
    service_type = service.replace("-", "_")
    path = context.service_mirror_cfg[service_type]["path"]
    cmd = "nohup sh -c 'python3 -m http.server --directory {} {} > /dev/null 2>&1 &'".format(  # noqa
        path, port
    )

    when_i_run_command(
        context,
        cmd,
        "with sudo",
        machine_name=machine_name,
    )

    if service_type in context.service_mirror_cfg:
        context.service_mirror_cfg[service_type]["port"] = port


@when(
    "I create the contract config overrides file for `{service_list}` on the `{machine_name}` machine"  # noqa
)
def create_contract_overrides(context, service_list, machine_name):
    token = context.pro_config.contract_token
    config_override = {token: {}}  # type: Dict[str, Any]

    for service in service_list.split(","):
        config_override[token][service] = {
            "directives": {
                "aptURL": "http://{}:{}".format(
                    context.machines[machine_name].instance.ip,
                    context.service_mirror_cfg.get(
                        service.replace("-", "_"), {}
                    ).get("port", "8000"),
                )
            }
        }

    context.text = yaml.dump(config_override)
    contract_override_path = "/tmp/contract-override"
    when_i_create_file_with_content(
        context,
        contract_override_path,
        machine_name=machine_name,
    )

    context.service_mirror_cfg["contract_override"] = contract_override_path


@when(
    "I generate the contracts-airgapped configuration on the `{machine_name}` machine"  # noqa
)
def i_configure_the_ua_airgapped_service(context, machine_name):
    contract_override_path = context.service_mirror_cfg["contract_override"]
    contract_final_cfg_path = "contract-server-ready.yml"
    cmd = "sh -c 'cat {} | pro-airgapped > {}'".format(
        contract_override_path,
        contract_final_cfg_path,
    )

    when_i_run_command(
        context,
        cmd,
        "with sudo",
        machine_name=machine_name,
    )

    context.service_mirror_cfg["contract_final_cfg"] = contract_final_cfg_path


@when(
    "I send the contracts-airgapped config from the `{base_machine}` machine to the `{target_machine}` machine"  # noqa
)
def i_fetch_contracts_airgapped_config(context, base_machine, target_machine):
    local_file_path = "/tmp/contracts-airgapped-cfg"

    context.machines[base_machine].instance.pull_file(
        context.service_mirror_cfg["contract_final_cfg"],
        local_file_path,
    )

    context.machines[target_machine].instance.push_file(
        local_file_path,
        context.service_mirror_cfg["contract_final_cfg"],
    )

    os.unlink(local_file_path)


@when(
    "I start the contracts-airgapped service on the `{machine_name}` machine"
)
def i_start_the_contracts_airgapped_service(context, machine_name):
    path = context.service_mirror_cfg["contract_final_cfg"]
    cmd = "nohup sh -c 'contracts-airgapped --input=./{} > /dev/null 2>&1 &'".format(  # noqa
        path
    )

    when_i_run_command(context, cmd, "with sudo", machine_name=machine_name)


@when(
    "I consolidate `{services_list}` on a single mirror on the `{machine_name}` machine"  # noqa
)  # noqa
def then_i_consolidate_services_on_the_same_mirror(
    context, services_list, machine_name
):
    services = services_list.split(",")
    all_mirrors_path = "/var/spool/apt-mirror/mirror/all-mirrors/"

    for service in services:
        cmd = "rsync -a /var/spool/apt-mirror/mirror/esm.ubuntu.com/{}/ {}".format(  # noqa
            service.split("-")[1], all_mirrors_path
        )
        when_i_run_command(
            context, cmd, "with sudo", machine_name=machine_name
        )

    context.service_mirror_cfg["all_mirrors"] = {"path": all_mirrors_path}



from typing import List, Optional

from uaclient.api.api import APIEndpoint
from uaclient.api.data_types import AdditionalInfo, ErrorWarningObject
from uaclient.api.u.pro.status.is_attached.v1 import _is_attached
from uaclient.config import UAConfig
from uaclient.data_types import (
    BoolDataValue,
    DataObject,
    Field,
    StringDataValue,
    data_list,
)


class EnabledService(DataObject):
    fields = [
        Field("name", StringDataValue, doc="Name of the service"),
        Field(
            "variant_enabled",
            BoolDataValue,
            doc="If a variant of the service is enabled",
        ),
        Field(
            "variant_name",
            StringDataValue,
            False,
            doc="Name of the variant, if a variant is enabled",
        ),
    ]

    def __init__(
        self,
        *,
        name: str,
        variant_enabled: bool = False,
        variant_name: Optional[str] = None
    ):
        self.name = name
        self.variant_enabled = variant_enabled
        self.variant_name = variant_name


class EnabledServicesResult(DataObject, AdditionalInfo):
    fields = [
        Field(
            "enabled_services",
            data_list(EnabledService),
            doc="A list of ``EnabledService`` objects",
        ),
    ]

    def __init__(self, *, enabled_services: List[EnabledService]):
        self.enabled_services = enabled_services


def enabled_services() -> EnabledServicesResult:
    return _enabled_services(UAConfig())


def _enabled_services(cfg: UAConfig) -> EnabledServicesResult:
    """
    This endpoint shows the Pro services that are enabled on the machine.
    """
    from uaclient.entitlements import ENTITLEMENT_CLASSES
    from uaclient.entitlements.entitlement_status import UserFacingStatus

    if not _is_attached(cfg).is_attached:
        return EnabledServicesResult(enabled_services=[])

    enabled_services = []  # type: List[EnabledService]
    warnings = []  # type: List[ErrorWarningObject]
    for ent_cls in ENTITLEMENT_CLASSES:
        ent = ent_cls(cfg)
        ent_status, details = ent.user_facing_status()

        if ent_status in (UserFacingStatus.ACTIVE, UserFacingStatus.WARNING):
            ent_name = ent.presentation_name
            enabled_service = EnabledService(name=ent_name)

            for _, variant_cls in ent.variants.items():
                variant = variant_cls(cfg)

                if variant.user_facing_status()[0] == UserFacingStatus.ACTIVE:
                    enabled_service = EnabledService(
                        name=ent_name,
                        variant_enabled=True,
                        variant_name=variant.variant_name,
                    )
                    break

            enabled_services.append(enabled_service)

            if ent_status == UserFacingStatus.WARNING and details:
                warnings.append(
                    ErrorWarningObject(
                        title=details.msg or "",
                        code=details.name or "",
                        meta={"service": ent_name},
                    )
                )

    result = EnabledServicesResult(
        enabled_services=sorted(enabled_services, key=lambda x: x.name),
    )
    result.warnings = warnings

    return result


endpoint = APIEndpoint(
    version="v1",
    name="EnabledServices",
    fn=_enabled_services,
    options_cls=None,
)

_doc = {
    "introduced_in": "28",
    "requires_network": False,
    "example_python": """
from uaclient.api.u.pro.status.enabled_services.v1 import enabled_services

result = enabled_services()
""",  # noqa: E501
    "result_class": EnabledServicesResult,
    "exceptions": [],
    "example_cli": "pro api u.pro.status.enabled_services.v1",
    "example_json": """
{
    "enabled_services": [
        {
            "name": "esm-apps",
            "variant_enabled": false,
            "variant_name": null
        },
        {
            "name": "esm-infra",
            "variant_enabled": false,
            "variant_name": null
        },
        {
            "name": "realtime-kernel",
            "variant_enabled": true,
            "variant_name": "raspi"
        }
    ]
}
""",
}



import mock

from uaclient.cli.system import reboot_required_subcommand

M_PATH = "uaclient.cli.system."


class TestActionRebootRequired:
    @mock.patch(M_PATH + "_reboot_required")
    def test_returns_the_api_response(self, m_api_reboot_required, event):
        m_cfg = mock.MagicMock()
        with mock.patch.object(event, "info") as m_event_info:
            reboot_required_subcommand.action(mock.MagicMock(), cfg=m_cfg)

        assert [mock.call(m_cfg)] == m_api_reboot_required.call_args_list
        assert [
            mock.call(m_api_reboot_required.return_value.reboot_required)
        ] == m_event_info.call_args_list



import argparse
from typing import Callable, Iterable, Optional, Union

from uaclient import messages
from uaclient.cli.parser import HelpCategory


class ProArgument:
    def __init__(
        self,
        long_name: str,
        help: str,
        short_name: Optional[str] = None,
        **kwargs
    ):
        self.names = (
            (long_name,) if short_name is None else (short_name, long_name)
        )
        self.help = help
        self.additional_args = kwargs

    def register(
        self, target: Union[argparse.ArgumentParser, argparse._ArgumentGroup]
    ):
        target.add_argument(
            *self.names, help=self.help, **self.additional_args
        )


class ProArgumentMutuallyExclusiveGroup:
    def __init__(
        self,
        required: bool = False,
        arguments: Iterable[ProArgument] = (),
    ):
        self.required = required
        self.arguments = arguments


class ProArgumentGroup:
    def __init__(
        self,
        title: Optional[str] = None,
        description: Optional[str] = None,
        arguments: Iterable[ProArgument] = (),
        mutually_exclusive_groups: Iterable[
            ProArgumentMutuallyExclusiveGroup
        ] = (),
    ):
        self.title = title
        self.description = description
        self.arguments = arguments
        self.mutually_exclusive_groups = mutually_exclusive_groups

    def register(self, parser: argparse.ArgumentParser):
        target = (
            parser
        )  # type: Union[argparse.ArgumentParser, argparse._ArgumentGroup]
        if self.title:
            target = parser.add_argument_group(self.title, self.description)

        for argument in self.arguments:
            argument.register(target)

        for group in self.mutually_exclusive_groups:
            new_group = target.add_mutually_exclusive_group(
                required=group.required
            )
            for argument in group.arguments:
                argument.register(new_group)


class ProCommand:
    def __init__(
        self,
        name: str,
        help: str,
        description: str,
        action: Callable = lambda *args, **kwargs: None,
        preserve_description: bool = False,
        argument_groups: Iterable[ProArgumentGroup] = (),
        subcommands: Iterable["ProCommand"] = (),
        help_category: Optional[HelpCategory] = None,
        help_position: int = 0,
    ):
        self.name = name
        self.help = help
        self.description = description
        self.action = action
        self.preserve_description = preserve_description
        self.argument_groups = argument_groups
        self.subcommands = subcommands
        self.help_category = help_category
        self.help_position = help_position

    def register(self, subparsers: argparse._SubParsersAction):
        self.parser = subparsers.add_parser(
            self.name,
            help=self.help,
            description=self.description,
        )
        if self.preserve_description:
            self.parser.formatter_class = argparse.RawDescriptionHelpFormatter

        if self.help_category:
            self.parser.add_help_entry(
                category=self.help_category,
                name=self.name,
                help_string=self.help,
                position=self.help_position,
            )

        for argument_group in self.argument_groups:
            argument_group.register(self.parser)

        self.parser.set_defaults(action=self.action)

        if self.subcommands:
            subparsers = self.parser.add_subparsers(
                title=messages.CLI_AVAILABLE_COMMANDS,
                dest="command",
            )
            for command in self.subcommands:
                command.register(subparsers)



import sys
import textwrap
from collections import namedtuple

from uaclient import defaults, exceptions, messages, system, util
from uaclient.api.u.pro.security.cves._common.v1 import VulnerabilityData
from uaclient.api.u.pro.security.cves.v1 import CVEInfo, CVEsOptions, _cves
from uaclient.api.u.pro.security.fix._common import (
    query_installed_source_pkg_versions,
)
from uaclient.cli import cli_util
from uaclient.cli.commands import ProArgument, ProArgumentGroup, ProCommand
from uaclient.cli.formatter import Table
from uaclient.cli.parser import HelpCategory
from uaclient.config import UAConfig

AffectedPackage = namedtuple(
    "AffectedPackage", ["name", "fix_status", "fix_origin", "fix_version"]
)


@cli_util.with_spinner(msg=messages.CLI_CVES_SPINNER_MSG)
def _get_cve_vulnerabilities(args, *, cfg: UAConfig, **kwargs):
    try:
        result = _cves(options=CVEsOptions(), cfg=cfg)
    except exceptions.VulnerabilityDataNotFound:
        result = None

    return result


def _get_affected_pkgs(cve_vulnerabilities, cve_info, cve_name):
    rows = []
    installed_pkgs_by_source = query_installed_source_pkg_versions()

    for source_pkg in cve_info.related_packages:
        binary_pkgs = installed_pkgs_by_source.get(source_pkg, {}).keys()

        for binary_pkg in sorted(binary_pkgs):
            binary_pkg_info = cve_vulnerabilities.packages.get(binary_pkg)

            if binary_pkg_info:
                for cve in binary_pkg_info.cves:
                    if cve.name == cve_name:
                        rows.append(
                            AffectedPackage(
                                name=binary_pkg,
                                fix_status=cve.fix_status,
                                fix_origin=cve.fix_origin,
                                fix_version=cve.fix_version,
                            )
                        )
                        break

    return rows


def _format_affected_pkgs(affected_pkgs):
    formatted_rows = []

    for affected_pkg in affected_pkgs:
        if affected_pkg.fix_status == "fixed":
            formatted_rows.append(
                [
                    "{}:".format(affected_pkg.name),
                    affected_pkg.fix_status,
                    "({})".format(affected_pkg.fix_origin),
                    affected_pkg.fix_version,
                ]
            )
        else:
            formatted_rows.append(
                [
                    "{}:".format(affected_pkg.name),
                    affected_pkg.fix_status,
                    "",
                    "",
                ]
            )

    return formatted_rows


def action_cve(args, *, cfg, **kwargs):
    cve_name = args.cve.upper()
    cve_vulnerabilities = _get_cve_vulnerabilities(args, cfg=cfg)

    if not cve_vulnerabilities:
        raise exceptions.VulnerabilityDataNotFound()

    if cve_name not in cve_vulnerabilities.cves:
        cve_data = (
            VulnerabilityData(cfg)
            .get()
            .get("security_issues", {})
            .get("cves", {})
            .get(cve_name)
        )

        if not cve_data:
            release = system.get_release_info().release
            print(
                messages.CLI_CVE_NOT_FOUND_IN_DATA.format(
                    issue=args.cve,
                    release=release,
                    url="{}/{}".format(defaults.BASE_SECURITY_URL, cve_name),
                ),
                file=sys.stderr,
            )
            return

        cve_info = CVEInfo(
            description=cve_data["description"],
            published_at=util.parse_rfc3339_date(cve_data["published_at"]),
            priority=cve_data["ubuntu_priority"],
            notes=cve_data["notes"],
            cvss_score=cve_data["cvss_score"],
            cvss_severity=cve_data["cvss_severity"],
            related_usns=[],
        )
        affected_pkgs_table = ""
    else:
        cve_info = cve_vulnerabilities.cves[cve_name]

        affected_pkgs_rows = _format_affected_pkgs(
            _get_affected_pkgs(cve_vulnerabilities, cve_info, args.cve)
        )
        affected_pkgs_table = Table(rows=affected_pkgs_rows).to_string()

    print("name:            {}".format(cve_name))
    print(
        "public-url:      {}/{}".format(defaults.BASE_SECURITY_URL, cve_name)
    )
    print(
        "published-at:    {}".format(
            cve_info.published_at.strftime("%Y-%m-%d")
        )
    )
    print(
        "cve-cache-date:  {}".format(
            cve_vulnerabilities.vulnerability_data_published_at.strftime(
                "%Y-%m-%d"
            )
        )
    )
    print(
        "apt-cache-date:  {}".format(
            cve_vulnerabilities.apt_updated_at.strftime("%Y-%m-%d")
        )
    )
    print(
        "priority:        {}".format(
            cli_util.colorize_priority(cve_info.priority)
        )
    )

    if cve_info.cvss_score:
        print("cvss-score:      {}".format(cve_info.cvss_score))

    if cve_info.cvss_severity:
        print("cvss-severity:   {}".format(cve_info.cvss_severity))

    print("description: |")
    print(
        "{}".format(
            "\n".join(
                textwrap.wrap(
                    cve_info.description,
                    width=defaults.PRINT_WRAP_WIDTH,
                    break_long_words=False,
                    break_on_hyphens=False,
                    initial_indent="  ",
                    subsequent_indent="  ",
                )
            )
        )
    )

    if cve_info.notes:
        print("notes:")
        for note in cve_info.notes:
            print(
                textwrap.fill(
                    note,
                    width=defaults.PRINT_WRAP_WIDTH,
                    break_long_words=False,
                    break_on_hyphens=False,
                    initial_indent="  - ",
                    subsequent_indent="    ",
                )
            )

    if affected_pkgs_table:
        print("affected_packages:")
        for line in affected_pkgs_table.splitlines():
            print("  " + line)
    else:
        print("affected_packages: []")

    if cve_info.related_usns:
        related_usns = [
            "  {}: {}".format(usn.name, usn.title)
            for usn in cve_info.related_usns
            if usn.title
        ]

        if related_usns:
            print("related_usns:")
            for related_usn in related_usns:
                print(related_usn)


cve_command = ProCommand(
    "cve",
    help=messages.CLI_CVE,
    description=messages.CLI_CVE_DESC,
    action=action_cve,
    help_category=HelpCategory.SECURITY,
    preserve_description=True,
    argument_groups=[
        ProArgumentGroup(
            arguments=[
                ProArgument(
                    "cve",
                    help=messages.CLI_CVE_ISSUE,
                ),
            ]
        )
    ],
)
